import logging

import h5py
import numpy as np

import westpa
from westpa.core.segment import Segment
from westpa.oldtools.aframe import AnalysisMixin
from westpa.oldtools.miscfn import parse_int_list

log = logging.getLogger(__name__)


class WESTDataReaderMixin(AnalysisMixin):
    '''A mixin for analysis requiring access to the HDF5 files generated during a WEST run.'''

    def __init__(self):
        super().__init__()

        self.data_manager = None
        self.west_h5name = None

        # Whether pcoord caching is active
        self.__cache_pcoords = False

        # Cached items
        self.__c_summary = None
        self.__c_iter_groups = dict()
        self.__c_seg_id_ranges = dict()
        self.__c_seg_indices = dict()
        self.__c_wtg_parent_arrays = dict()
        self.__c_parent_arrays = dict()
        self.__c_pcoord_arrays = dict()
        self.__c_pcoord_datasets = dict()

    def add_args(self, parser, upcall=True):
        if upcall:
            try:
                upcall = super().add_args
            except AttributeError:
                pass
            else:
                upcall(parser)

        group = parser.add_argument_group('WEST input data options')
        group.add_argument(
            '-W',
            '--west-data',
            dest='west_h5name',
            metavar='WEST_H5FILE',
            help='''Take WEST data from WEST_H5FILE (default: read from the HDF5 file specified in west.cfg).''',
        )

    def process_args(self, args, upcall=True):
        if args.west_h5name:
            self.west_h5name = args.west_h5name
        else:
            westpa.rc.config.require(['west', 'data', 'west_data_file'])
            self.west_h5name = westpa.rc.config.get_path(['west', 'data', 'west_data_file'])

        westpa.rc.pstatus("Using WEST data from '{}'".format(self.west_h5name))

        self.data_manager = westpa.rc.get_data_manager()
        self.data_manager.we_h5filename = self.west_h5name
        self.data_manager.open_backing(mode='r')

        if upcall:
            try:
                upfunc = super().process_args
            except AttributeError:
                pass
            else:
                upfunc(args)

    def clear_run_cache(self):
        del self.__c_summary
        del self.__c_iter_groups, self.__c_seg_id_ranges, self.__c_seg_indices, self.__c_parent_arrays, self.__c_parent_arrays
        del self.__c_pcoord_arrays, self.__c_pcoord_datasets

        self.__c_summary = None
        self.__c_iter_groups = dict()
        self.__c_seg_id_ranges = dict()
        self.__c_seg_indices = dict()
        self.__c_parent_arrays = dict()
        self.__c_wtg_parent_arrays = dict()
        self.__c_pcoord_arrays = dict()
        self.__c_pcoord_datasets = dict()

    @property
    def cache_pcoords(self):
        '''Whether or not to cache progress coordinate data. While caching this data
        can significantly speed up some analysis operations, this requires
        copious RAM.

        Setting this to False when it was formerly True will release any cached data.
        '''
        return self.__cache_pcoords

    @cache_pcoords.setter
    def cache_pcoords(self, cache):
        self.__cache_pcoords = cache

        if not cache:
            del self.__c_pcoord_arrays
            self.__c_pcoord_arrays = dict()

    def get_summary_table(self):
        if self.__c_summary is None:
            self.__c_summary = self.data_manager.we_h5file['/summary'][...]
        return self.__c_summary

    def get_iter_group(self, n_iter):
        '''Return the HDF5 group corresponding to ``n_iter``'''
        try:
            return self.__c_iter_groups[n_iter]
        except KeyError:
            iter_group = self.data_manager.get_iter_group(n_iter)
            return iter_group

    def get_segments(self, n_iter, include_pcoords=True):
        '''Return all segments present in iteration n_iter'''
        return self.get_segments_by_id(n_iter, self.get_seg_ids(n_iter, None), include_pcoords)

    def get_segments_by_id(self, n_iter, seg_ids, include_pcoords=True):
        '''Get segments from the data manager, employing caching where possible'''

        if len(seg_ids) == 0:
            return []

        seg_index = self.get_seg_index(n_iter)
        all_wtg_parent_ids = self.get_wtg_parent_array(n_iter)

        segments = []

        if include_pcoords:
            pcoords = self.get_pcoords(n_iter, seg_ids)

        for isegid, seg_id in enumerate(seg_ids):
            row = seg_index[seg_id]
            parents_offset = row['wtg_offset']
            n_parents = row['wtg_n_parents']
            segment = Segment(
                seg_id=seg_id,
                n_iter=n_iter,
                status=row['status'],
                endpoint_type=row['endpoint_type'],
                walltime=row['walltime'],
                cputime=row['cputime'],
                weight=row['weight'],
            )
            if include_pcoords:
                segment.pcoord = pcoords[isegid]

            parent_ids = all_wtg_parent_ids[parents_offset : parents_offset + n_parents]
            segment.wtg_parent_ids = {int(parent_id) for parent_id in parent_ids}
            segment.parent_id = int(parent_ids[0])
            segments.append(segment)

        return segments

    def get_children(self, segment, include_pcoords=True):
        parents = self.get_parent_array(segment.n_iter + 1)
        seg_ids = self.get_seg_ids(segment.n_iter + 1, parents == segment.seg_id)
        return self.get_segments_by_id(segment.n_iter + 1, seg_ids, include_pcoords)

    def get_seg_index(self, n_iter):
        try:
            return self.__c_seg_indices[n_iter]
        except KeyError:
            seg_index = self.__c_seg_indices[n_iter] = self.get_iter_group(n_iter)['seg_index'][...]
            return seg_index

    def get_wtg_parent_array(self, n_iter):
        try:
            return self.__c_wtg_parent_arrays[n_iter]
        except KeyError:
            parent_array = self.__c_wtg_parent_arrays[n_iter] = self.get_iter_group(n_iter)['wtgraph'][...]
            return parent_array

    def get_parent_array(self, n_iter):
        try:
            return self.__c_parent_arrays[n_iter]
        except KeyError:
            parent_array = self.get_seg_index(n_iter)['parent_id']
            self.__c_parent_arrays[n_iter] = parent_array
            return parent_array

    def get_pcoord_array(self, n_iter):
        try:
            return self.__c_pcoord_arrays[n_iter]
        except KeyError:
            pcoords = self.__c_pcoord_arrays[n_iter] = self.get_iter_group(n_iter)['pcoord'][...]
            return pcoords

    def get_pcoord_dataset(self, n_iter):
        try:
            return self.__c_pcoord_datasets[n_iter]
        except KeyError:
            pcoord_ds = self.__c_pcoord_datasets[n_iter] = self.get_iter_group(n_iter)['pcoord']
            return pcoord_ds

    def get_pcoords(self, n_iter, seg_ids):
        if self.__cache_pcoords:
            pcarray = self.get_pcoord_array(n_iter)
            return [pcarray[seg_id, ...] for seg_id in seg_ids]
        else:
            return self.get_pcoord_dataset(n_iter)[list(seg_ids), ...]

    def get_seg_ids(self, n_iter, bool_array=None):
        try:
            all_ids = self.__c_seg_id_ranges[n_iter]
        except KeyError:
            all_ids = self.__c_seg_id_ranges[n_iter] = np.arange(0, len(self.get_seg_index(n_iter)), dtype=np.uint32)

        if bool_array is None:
            return all_ids
        else:
            seg_ids = all_ids[bool_array]
            try:
                if len(seg_ids) == 0:
                    return []
            except TypeError:
                # Not iterable, for some bizarre reason
                return [seg_ids]
            else:
                return seg_ids

    def get_created_seg_ids(self, n_iter):
        '''Return a list of seg_ids corresponding to segments which were created for the given iteration (are not
        continuations).'''

        # Created segments have parent_id < 0
        parent_ids = self.get_parent_array(n_iter)
        return self.get_seg_ids(n_iter, parent_ids < 0)

    def max_iter_segs_in_range(self, first_iter, last_iter):
        '''Return the maximum number of segments present in any iteration in the range selected'''
        n_particles = self.get_summary_table()['n_particles']
        return n_particles[first_iter - 1 : last_iter].max()

    def total_segs_in_range(self, first_iter, last_iter):
        '''Return the total number of segments present in all iterations in the range selected'''
        n_particles = self.get_summary_table()['n_particles']
        return n_particles[first_iter - 1 : last_iter].sum()

    def get_pcoord_len(self, n_iter):
        '''Get the length of the progress coordinate array for the given iteration.'''
        pcoord_ds = self.get_pcoord_dataset(n_iter)
        return pcoord_ds.shape[1]

    def get_total_time(self, first_iter=None, last_iter=None, dt=None):
        '''Return the total amount of simulation time spanned between first_iter and last_iter (inclusive).'''
        first_iter = first_iter or self.first_iter
        last_iter = last_iter or self.last_iter
        dt = dt or getattr(self, 'dt', 1.0)

        total_len = 0
        for n_iter in range(first_iter, last_iter + 1):
            total_len += self.get_pcoord_len(n_iter) - 1
        return total_len * dt


class ExtDataReaderMixin(AnalysisMixin):
    '''An external data reader, primarily designed for reading brute force data, but also suitable
    for any auxiliary datasets required for analysis.
    '''

    default_chunksize = 8192

    def __init__(self):
        super().__init__()

        self.ext_input_nargs = '+'
        self.ext_input_filenames = []
        self.ext_input_chunksize = self.default_chunksize
        self.ext_input_usecols = None
        self.ext_input_comment_regexp = None
        self.ext_input_sep_regexp = None

    def add_args(self, parser, upcall=True):
        if upcall:
            try:
                upcall = super().add_args
            except AttributeError:
                pass
            else:
                upcall(parser)

        input_options = parser.add_argument_group('external data input options')
        input_options.add_argument(
            'datafiles',
            nargs=self.ext_input_nargs,
            metavar='DATAFILE',
            help='''Data file(s) to analyze, either text or Numpy (.npy or .npz) format.
                                   Uncompressed numpy files will be memory-mapped, allowing analysis of data larger than
                                   available RAM (though not larger than the available address space).''',
        )
        input_options.add_argument(
            '--usecols',
            dest='usecols',
            metavar='COLUMNS',
            type=parse_int_list,
            help='''Use only the given COLUMNS from the input file(s), e.g. "0", "0,1",
                                   "0:5,7,9:10".''',
        )
        input_options.add_argument(
            '--chunksize',
            dest='chunksize',
            type=int,
            default=self.default_chunksize,
            help='''Process input data in blocks of size CHUNKSIZE.  This will only reduce memory
                                   requirements when using uncompressed Numpy (.npy) format input. (Default: %(default)d.)''',
        )

    def process_args(self, args, upcall=True):
        if args.usecols:
            westpa.rc.pstatus('Using only the following columns from external input: {!s}'.format(args.usecols))
            self.ext_input_usecols = args.usecols
        else:
            self.ext_input_usecols = None

        self.ext_input_filenames = args.datafiles
        self.ext_input_chunksize = args.chunksize or self.default_chunksize

        if upcall:
            try:
                upfunc = super().process_args
            except AttributeError:
                pass
            else:
                upfunc(args)

    def is_npy(self, filename):
        with open(filename, 'rb') as fileobj:
            first_bytes = fileobj.read(len(np.lib.format.MAGIC_PREFIX))

        if first_bytes == np.lib.format.MAGIC_PREFIX:
            return True
        else:
            return False

    def load_npy_or_text(self, filename):
        '''Load an array from an existing .npy file, or read a text file and
        convert to a NumPy array.  In either case, return a NumPy array.  If a
        pickled NumPy dataset is found, memory-map it read-only.  If the specified
        file does not contain a pickled NumPy array, attempt to read the file using
        numpy.loadtxt(filename).'''

        if self.is_npy(filename):
            return np.load(filename, 'r')
        else:
            return np.loadtxt(filename)

    def text_to_h5dataset(self, fileobj, group, dsname, dtype=np.float64, skiprows=0, usecols=None, chunksize=None):
        '''Read text-format data from the given filename or file-like object ``fileobj`` and write to a newly-created dataset
        called ``dsname`` in the HDF5 group ``group``.  The data is stored as type ``dtype``.  By default, the shape is
        taken as (number of lines, number of columns); columns can be omitted by specifying a list for ``usecols``,
        and lines can be skipped by using ``skiprows``.  Data is read in chunks of ``chunksize`` rows.'''

        try:
            fileobj.readline
        except AttributeError:
            fileobj = open(fileobj, 'rt')

        usecols = usecols or self.usecols
        chunksize = chunksize or self.ext_input_chunksize

        linenumber = 0
        for iskip in range(skiprows or 0):
            fileobj.readline()
            linenumber += 1

        nrows = 0
        irow = 0
        ncols_input = None  # number of columns in input
        ncols_store = None  # number of columns to store
        databuffer = None
        dataset = None

        re_split_comments = self.ext_input_comment_regexp
        re_split_fields = self.ext_input_sep_regexp

        for line in fileobj:
            linenumber += 1

            # Discard comments and extraneous whitespace
            if re_split_comments is not None:
                record_text = re_split_comments.split(line, 1)[0].strip()
            else:
                record_text = line.split('#', 1)[0].strip()

            if not record_text:
                continue

            if re_split_fields is not None:
                fields = re_split_fields.split(record_text)
            else:
                fields = record_text.split()

            # Check that the input size hasn't change (blank lines excluded)
            if not ncols_input:
                ncols_input = len(fields)
            elif len(fields) != ncols_input:
                raise ValueError('expected {:d} columns at line {:d}, but found {:d}'.format(ncols_input, linenumber, len(fields)))

            # If this is the first time through the loop, allocate temporary storage
            if not ncols_store:
                ncols_store = len(usecols)
                databuffer = np.empty((chunksize, ncols_store), dtype)
                dataset = group.create_dataset(
                    dsname, shape=(0, ncols_store), maxshape=(None, ncols_store), chunks=(chunksize, ncols_store), dtype=dtype
                )

            if usecols:
                for ifield, iifield in enumerate(usecols):
                    databuffer[irow, ifield] = dtype(fields[iifield])
            else:
                for ifield, field in enumerate(fields):
                    databuffer[irow, ifield] = dtype(field)

            nrows += 1
            irow += 1

            # Flush to HDF5 if necessary
            if irow == chunksize:
                westpa.rc.pstatus('\r  Read {:d} rows'.format(nrows), end='')
                westpa.rc.pflush()
                dataset.resize((nrows, ncols_store))
                dataset[-irow:] = databuffer
                irow = 0

        # Flush last bit
        if irow > 0:
            dataset.resize((nrows, ncols_store))
            dataset[-irow:] = databuffer[:irow]
        westpa.rc.pstatus('\r  Read {:d} rows'.format(nrows))
        westpa.rc.pflush()

    def npy_to_h5dataset(self, array, group, dsname, usecols=None, chunksize=None):
        '''Store the given array into a newly-created dataset named ``dsname`` in the HDF5 group
        ``group``, optionally only storing a subset of columns.  Data is written ``chunksize`` rows at a time,
        allowing very large memory-mapped arrays to be copied.'''

        usecols = usecols or self.ext_input_usecols
        chunksize = chunksize or self.ext_input_chunksize

        if usecols:
            shape = (len(array),) + array[0][usecols].shape[1:]
        else:
            shape = array.shape

        if len(shape) == 1:
            shape = shape + (1,)
        maxlen = len(array)
        mw = len(str(maxlen))
        dataset = group.create_dataset(dsname, shape=shape, dtype=array.dtype)

        if usecols:
            for istart in range(0, maxlen, chunksize):
                iend = min(istart + chunksize, maxlen)
                dataset[istart:iend] = array[istart:iend, usecols]
                westpa.rc.pstatus('\r  Read {:{mw}d}/{:>{mw}d} rows'.format(iend, maxlen, mw=mw), end='')
                westpa.rc.pflush()
        else:
            for istart in range(0, maxlen, chunksize):
                dataset[istart:iend] = array[istart:iend]
                westpa.rc.pstatus('\r  Read {:{mw}d}/{:>{mw}d} rows'.format(iend, maxlen, mw=mw), end='')
                westpa.rc.pflush()
        westpa.rc.pstatus()


class BFDataManager(AnalysisMixin):
    '''A class to manage brute force trajectory data.  The primary purpose is to read in and
    manage brute force progress coordinate data for one or more trajectories.  The trajectories need not
    be the same length, but they do need to have the same time spacing for progress coordinate values.'''

    traj_index_dtype = np.dtype([('pcoord_len', np.uint64), ('source_data', h5py.special_dtype(vlen=str))])

    def __init__(self):
        super().__init__()
        self.bf_h5name = None
        self.bf_h5file = None

    def add_args(self, parser, upcall=True):
        if upcall:
            try:
                upcall = super().add_args
            except AttributeError:
                pass
            else:
                upcall(parser)

        group = parser.add_argument_group('brute force input data options')
        group.add_argument(
            '-B',
            '--bfdata',
            '--brute-force-data',
            dest='bf_h5name',
            metavar='BF_H5FILE',
            default='bf_system.h5',
            help='''Brute force data is/will be stored in BF_H5FILE (default: %(default)s).''',
        )

    def process_args(self, args, upcall=True):
        self.bf_h5name = args.bf_h5name
        westpa.rc.pstatus("Using brute force data from '{}'".format(self.bf_h5name))

        if upcall:
            try:
                upfunc = super().process_args
            except AttributeError:
                pass
            else:
                upfunc(args)

    def _get_traj_group_name(self, traj_id):
        return 'traj_{:09d}'.format(traj_id)

    def update_traj_index(self, traj_id, pcoord_len, source_data):
        self.bf_h5file['traj_index'][traj_id] = (pcoord_len, source_data)

    def get_traj_group(self, traj_id):
        return self.bf_h5file[self._get_traj_group_name(traj_id)]

    def create_traj_group(self):
        new_traj_id = self.get_n_trajs()
        group = self.bf_h5file.create_group(self._get_traj_group_name(new_traj_id))
        self.bf_h5file['traj_index'].resize((new_traj_id + 1,))
        return (new_traj_id, group)

    def get_n_trajs(self):
        return self.bf_h5file['traj_index'].shape[0]

    def get_traj_len(self, traj_id):
        return self.bf_h5file['traj_index'][traj_id]['pcoord_len']

    def get_max_traj_len(self):
        return self.bf_h5file['traj_index']['pcoord_len'].max()

    def get_pcoord_array(self, traj_id):
        return self.get_traj_group(traj_id)['pcoord'][...]

    def get_pcoord_dataset(self, traj_id):
        return self.get_traj_group(traj_id)['pcoord']

    def require_bf_h5file(self):
        if self.bf_h5file is None:
            assert self.bf_h5name
            self.bf_h5file = h5py.File(self.bf_h5name)
            try:
                self.bf_h5file['traj_index']
            except KeyError:
                # A new file; create the trajectory index
                self.bf_h5file.create_dataset('traj_index', shape=(0,), maxshape=(None,), dtype=self.traj_index_dtype)
        return self.bf_h5file

    def close_bf_h5file(self):
        if self.bf_h5file is not None:
            self.bf_h5file.close()
            self.bf_h5file = None



import logging

import numpy as np

import westpa
from westpa.oldtools.aframe import AnalysisMixin
from westpa.oldtools.aframe.trajwalker import TrajWalker

log = logging.getLogger(__name__)


class TransitionEventAccumulator:
    index_dtype = np.uintp
    count_dtype = np.uint64
    weight_dtype = np.float64
    output_tdat_chunksize = 4096  # HDF5 chunksize for transition data (~300 KiB)
    tdat_buffersize = 524288  # Internal buffer length (~38 MiB)
    max_acc = 32768

    def __init__(self, n_bins, output_group, calc_fpts=True):
        self.calc_fpts = calc_fpts
        self.n_bins = n_bins
        self.iibins = np.arange(n_bins)
        self.iibdisc = np.empty((n_bins,), np.bool_)

        self.bin_index_dtype = np.min_scalar_type(n_bins)

        self.tdat_dtype = np.dtype(
            [
                ('traj', self.index_dtype),
                ('n_iter', self.index_dtype),
                ('timepoint', self.index_dtype),
                ('initial_bin', self.bin_index_dtype),
                ('final_bin', self.bin_index_dtype),
                ('initial_weight', self.weight_dtype),
                ('final_weight', self.weight_dtype),
                ('initial_bin_pop', self.weight_dtype),
                ('duration', self.index_dtype),
                ('fpt', self.index_dtype),
            ]
        )

        # HDF5 group in which to store results
        self.output_group = output_group
        self.tdat_buffer = np.empty((self.tdat_buffersize,), dtype=self.tdat_dtype)
        self.tdat_buffer_offset = 0
        self.output_tdat_offset = 0
        self.output_tdat_ds = None

        # Accumulators/counters
        self.n_trans = None  # shape (n_bins,n_bins)

        # Time points and per-timepoint data
        self.last_exit = None  # (n_bins,)
        self.last_entry = None  # (n_bins,)
        self.last_completion = None  # (n_bins,n_bins)
        self.weight_last_exit = None  # (n_bins)
        self.bin_pops_last_exit = None  # (n_bins,)

        # Analysis continuation information
        self.timepoint = None  # current time index for separate calls on same trajectory
        self.last_bin = None  # last region occupied, for separate calls on same trajectory
        self.last_bin_pop = None  # total weight in self.last_region at end of last processing step

        self.clear()

    def clear(self):
        self.clear_state()
        self.n_trans = np.zeros((self.n_bins, self.n_bins), self.count_dtype)
        self.tdat_buffer = np.empty((self.tdat_buffersize,), dtype=self.tdat_dtype)
        self.tdat_buffer_offset = 0
        self.output_tdat_offset = 0
        self.output_tdat_ds = None

    def clear_state(self):
        self.last_exit = np.zeros((self.n_bins,), self.index_dtype)
        self.last_entry = np.zeros((self.n_bins,), self.index_dtype)
        self.last_completion = np.zeros((self.n_bins, self.n_bins), self.index_dtype)
        self.weight_last_exit = np.zeros((self.n_bins,), self.weight_dtype)
        self.bin_pops_last_exit = np.zeros((self.n_bins,), self.weight_dtype)
        self.timepoint = 0
        self.last_bin = None
        self.last_bin_pop = None

    def get_state(self):
        return {
            'last_entry': self.last_entry.copy(),
            'last_exit': self.last_exit.copy(),
            'last_completion': self.last_completion.copy(),
            'weight_last_exit': self.weight_last_exit.copy(),
            'bin_pops_last_exit': self.bin_pops_last_exit.copy(),
            'timepoint': self.timepoint,
            'last_bin': self.last_bin,
            'last_bin_pop': self.last_bin_pop,
        }

    def set_state(self, state_dict):
        self.last_entry = state_dict['last_entry']
        self.last_exit = state_dict['last_exit']
        self.last_completion = state_dict['last_completion']
        self.weight_last_exit = state_dict['weight_last_exit']
        self.bin_pops_last_exit = state_dict['bin_pops_last_exit']
        self.timepoint = state_dict['timepoint']
        self.last_bin = state_dict['last_bin']
        self.last_bin_pop = state_dict['last_bin_pop']

    def record_transition_data(self, tdat):
        """Update running statistics and write transition data to HDF5 (with buffering)"""

        # Write out accumulated transition data
        if self.output_tdat_ds is None:
            # Create dataset
            try:
                del self.output_group['transitions']
            except KeyError:
                pass

            self.output_tdat_ds = self.output_group.create_dataset(
                'transitions',
                shape=(1,),
                dtype=self.tdat_dtype,
                maxshape=(None,),
                chunks=(self.output_tdat_chunksize,),
                compression='gzip',
            )

        # If the amount of data to write exceeds our remaining buffer space, flush the buffer, then
        # write data directly to HDF5, otherwise just add to the buffer and wait for the last flush
        if len(tdat) + self.tdat_buffer_offset > self.tdat_buffersize:
            self.flush_transition_data()
            ub = self.output_tdat_offset + len(tdat)
            self.output_tdat_ds.resize((ub,))
            self.output_tdat_ds[self.output_tdat_offset : ub] = tdat
            self.output_tdat_offset += len(tdat)
        else:
            self.tdat_buffer[self.tdat_buffer_offset : (self.tdat_buffer_offset + len(tdat))] = tdat
            self.tdat_buffer_offset += len(tdat)

    def flush_transition_data(self):
        """Flush any unwritten output that may be present"""
        if self.output_tdat_ds is None:
            return

        # self.tdat_buffer_offset is the number of items in the buffer
        nbuf = self.tdat_buffer_offset
        if nbuf == 0:
            return
        ub = nbuf + self.output_tdat_offset
        if ub > self.output_tdat_ds.len():
            # Resize dataset to fit data
            self.output_tdat_ds.resize((ub,))

        self.output_tdat_ds[self.output_tdat_offset : ub] = self.tdat_buffer[:nbuf]
        self.output_tdat_offset += nbuf
        self.tdat_buffer_offset = 0

    def start_accumulation(self, assignments, weights, bin_pops, traj=0, n_iter=0):
        self.clear_state()
        timepoints = np.arange(len(assignments))
        self._accumulate_transitions(timepoints, assignments, weights, bin_pops, traj, n_iter)

    def continue_accumulation(self, assignments, weights, bin_pops, traj=0, n_iter=0):
        aug_assign = np.empty((len(assignments) + 1,), assignments.dtype)
        aug_assign[0] = self.last_bin
        aug_assign[1:] = assignments

        aug_weights = np.empty((len(weights) + 1,), self.weight_dtype)
        aug_weights[0] = 0
        aug_weights[1:] = weights

        aug_pops = np.empty((len(bin_pops) + 1, len(bin_pops[0])), self.weight_dtype)
        aug_pops[0, :] = 0
        aug_pops[0, self.last_bin] = self.last_bin_pop
        aug_pops[1:] = bin_pops

        timepoints = np.arange(self.timepoint, self.timepoint + len(aug_assign))
        self._accumulate_transitions(timepoints, aug_assign, aug_weights, aug_pops, traj, n_iter)

    def _accumulate_transitions(self, timepoints, assignments, weights, bin_pops, traj, n_iter):
        tdat = []

        assignments_from_1 = assignments[1:]
        assignments_to_1 = assignments[:-1]

        calc_fpts = self.calc_fpts

        trans_occur = assignments_from_1 != assignments_to_1
        trans_ibin = assignments_to_1[trans_occur]
        trans_fbin = assignments_from_1[trans_occur]
        trans_timepoints = timepoints[1:][trans_occur]
        trans_weights = weights[1:][trans_occur]  # arrival weights
        trans_ibinpops = bin_pops[:-1][trans_occur]

        last_exit = self.last_exit
        last_entry = self.last_entry
        last_completion = self.last_completion
        bin_pops_last_exit = self.bin_pops_last_exit
        weight_last_exit = self.weight_last_exit
        n_trans = self.n_trans
        iibdisc = self.iibdisc
        iibins = self.iibins
        tdat_maxlen = self.max_acc
        for trans_ti, weight, ibin, fbin, ibinpops in zip(trans_timepoints, trans_weights, trans_ibin, trans_fbin, trans_ibinpops):
            # Record this crossing event's data
            bin_pops_last_exit[ibin] = ibinpops[ibin]
            last_exit[ibin] = trans_ti
            last_entry[fbin] = trans_ti
            weight_last_exit[ibin] = weight

            # See what other transitions this crossing event completes
            iibdisc[:] = last_exit > 0
            iibdisc &= last_entry > last_completion[:, fbin]

            # Calculate event durations, etc for each transition generated by this crossing event
            durations = -last_exit + trans_ti + 1  # = time now - time of exit from initial bin
            if calc_fpts:
                fpts = -last_completion[fbin, :] + trans_ti  # = time now - time of last final->initial transition
                fpts[last_completion[:, fbin] == 0] = 0

                for iibin in iibins[iibdisc]:
                    tdat.append(
                        (
                            traj,
                            n_iter,
                            trans_ti,
                            iibin,
                            fbin,
                            weight_last_exit[iibin],
                            weight,
                            bin_pops_last_exit[iibin],
                            durations[iibin],
                            fpts[iibin],
                        )
                    )
            else:
                for iibin in iibins[iibdisc]:
                    tdat.append(
                        (
                            traj,
                            n_iter,
                            trans_ti,
                            iibin,
                            fbin,
                            weight_last_exit[iibin],
                            weight,
                            bin_pops_last_exit[iibin],
                            durations[iibin],
                            0,
                        )
                    )

            # Update tracking and statistics
            last_completion[iibdisc, fbin] = trans_ti
            n_trans[iibdisc, fbin] += 1

            if len(tdat) > tdat_maxlen:
                self.record_transition_data(tdat)

        self.record_transition_data(tdat)
        self.timepoint = timepoints[-1]
        self.last_bin = assignments[-1]
        self.last_bin_pop = bin_pops[-1, assignments[-1]]


class TransitionAnalysisMixin(AnalysisMixin):
    def __init__(self):
        super().__init__()
        self.discard_transition_data = False
        self.calc_fpts = False
        self.trans_h5gname = 'transitions'
        self.trans_h5group = None
        self.__transitions_ds = None
        self.n_trajs = 0

    def require_transitions_group(self):
        if self.trans_h5group is None:
            self.trans_h5group = self.anal_h5file.require_group(self.trans_h5gname)
        return self.trans_h5group

    def delete_transitions_group(self):
        self.trans_h5group = None
        del self.anal_h5file[self.trans_h5gname]

    def get_transitions_ds(self):
        if self.__transitions_ds is not None:
            return self.__transitions_ds
        else:
            self.__transitions_ds = self.trans_h5group['transitions']
            return self.__transitions_ds

    def add_args(self, parser, upcall=True):
        if upcall:
            try:
                upfunc = super().add_args
            except AttributeError:
                pass
            else:
                upfunc(parser)

        group = parser.add_argument_group('transition analysis options')
        group.add_argument(
            '--discard-transition-data',
            dest='discard_transition_data',
            action='store_true',
            help='''Discard any existing transition data stored in the analysis HDF5 file.''',
        )

    def process_args(self, args, upcall=True):
        self.discard_transition_data = args.discard_transition_data

        if upcall:
            try:
                upfunc = super().process_args
            except AttributeError:
                pass
            else:
                upfunc(args)

    def require_transitions(self):
        self.require_bin_assignments()
        self.require_transitions_group()
        do_trans = False
        if self.discard_transition_data:
            westpa.rc.pstatus('Discarding existing transition data.')
            do_trans = True
        elif not self.check_data_binhash(self.trans_h5group):
            westpa.rc.pstatus('Bin definitions have changed; deleting existing transition data.')
            do_trans = True
        elif 'transitions' in self.trans_h5group and not self.check_data_iter_range_least(self.trans_h5group):
            westpa.rc.pstatus('Existing transition data is for different first/last iterations; deleting.')
            do_trans = True

        if do_trans:
            self.delete_transitions_group()
            self.find_transitions()

    def find_transitions(self):
        westpa.rc.pstatus('Finding transitions...')
        output_group = self.require_transitions_group()

        self.n_segs_visited = 0
        self.n_total_segs = self.total_segs_in_range(self.first_iter, self.last_iter)
        self.accumulator = TransitionEventAccumulator(self.n_bins, output_group, calc_fpts=self.calc_fpts)
        self.bin_assignments = self.get_bin_assignments(self.first_iter, self.last_iter)
        self.bin_populations = self.get_bin_populations(self.first_iter, self.last_iter)

        walker = TrajWalker(data_reader=self)

        self.__pcoord_len = self.get_pcoord_len(self.first_iter)
        self.__quiet_mode = westpa.rc.quiet_mode

        walker.trace_trajectories(
            self.first_iter,
            self.last_iter,
            callable=self._segment_callback,
            include_pcoords=False,
            get_state=self.accumulator.get_state,
            set_state=self.accumulator.set_state,
        )
        self.accumulator.flush_transition_data()
        try:
            del output_group['n_trans']
        except KeyError:
            pass
        output_group['n_trans'] = self.accumulator.n_trans

        for h5object in (output_group, output_group['n_trans'], output_group['transitions']):
            self.record_data_iter_range(h5object)
            self.record_data_binhash(h5object)
            h5object.attrs['n_trajs'] = self.n_trajs
            h5object.attrs['n_segs'] = self.n_segs_visited

        self.accumulator.clear()
        westpa.rc.pstatus()

    def _segment_callback(self, segment, children, history):
        iiter = segment.n_iter - self.first_iter
        seg_id = segment.seg_id
        weights = np.empty((self.__pcoord_len,), np.float64)
        weights[:] = segment.weight
        bin_pops = self.bin_populations[iiter, :, :]

        if len(history) == 0:
            # New trajectory
            self.n_trajs += 1
            self.accumulator.start_accumulation(
                self.bin_assignments[iiter, seg_id, :], weights, bin_pops, traj=self.n_trajs, n_iter=segment.n_iter
            )
        else:
            # Continuing trajectory
            self.accumulator.continue_accumulation(
                self.bin_assignments[iiter, seg_id, :], weights, bin_pops, traj=self.n_trajs, n_iter=segment.n_iter
            )

        self.n_segs_visited += 1

        if not self.__quiet_mode and (self.n_segs_visited % 1000 == 0 or self.n_segs_visited == self.n_total_segs):
            pct_visited = self.n_segs_visited / self.n_total_segs * 100
            westpa.rc.pstatus(
                '\r  {:d} of {:d} segments ({:.1f}%) analyzed ({:d} independent trajectories)'.format(
                    int(self.n_segs_visited), int(self.n_total_segs), float(pct_visited), self.n_trajs
                ),
                end='',
            )
            westpa.rc.pflush()


class BFTransitionAnalysisMixin(TransitionAnalysisMixin):
    def require_transitions(self):
        self.require_bin_assignments()
        self.require_transitions_group()
        do_trans = False
        if self.discard_transition_data:
            westpa.rc.pstatus('Discarding existing transition data.')
            do_trans = True
        elif not self.check_data_binhash(self.trans_h5group):
            westpa.rc.pstatus('Bin definitions have changed; deleting existing transition data.')
            do_trans = True

        if do_trans:
            self.delete_transitions_group()
            self.find_transitions()

    def find_transitions(self, chunksize=65536):
        self.require_bf_h5file()
        self.require_binning_group()
        westpa.rc.pstatus('Finding transitions...')
        output_group = self.require_analysis_group('transitions')

        self.accumulator = TransitionEventAccumulator(self.n_bins, output_group, calc_fpts=True)
        assignments_ds = self.binning_h5group['bin_assignments']

        max_nrows = assignments_ds.len()
        maxwidth_nrows = len(str(max_nrows))

        for traj_id in range(self.get_n_trajs()):
            nrows = self.get_traj_len(traj_id)
            for istart in range(0, nrows, chunksize):
                iend = min(istart + chunksize, nrows)
                assignments = assignments_ds[traj_id, istart:iend]
                weights = np.ones((len(assignments),))
                binpops = np.ones((len(assignments), self.n_bins))

                if istart == 0:
                    self.accumulator.start_accumulation(assignments, weights, binpops, traj=traj_id)
                else:
                    self.accumulator.continue_accumulation(assignments, weights, binpops, traj=traj_id)

                westpa.rc.pstatus(
                    '\r  Trajectory {:d}: {:{mwnr}d}/{:<{mwnr}d} ({:.2f}%)'.format(
                        int(traj_id), int(iend), int(nrows), iend / nrows * 100, mwnr=maxwidth_nrows
                    ),
                    end='',
                )
                westpa.rc.pflush()
                self.accumulator.flush_transition_data()
                del assignments, weights, binpops
            westpa.rc.pstatus()

        try:
            del output_group['n_trans']
        except KeyError:
            pass
        output_group['n_trans'] = self.accumulator.n_trans

        for h5object in (output_group, output_group['n_trans'], output_group['transitions']):
            self.record_data_binhash(h5object)

        self.accumulator.clear()



import numpy as np


def time_average(observable, iterations):
    """Compute the time average of an observable.

    Parameters
    ----------
    observable : Callable[[Walker], ArrayLike]
        Function that takes a walker as input and returns a number or
        a fixed-size array of numbers.
    iterations : Sequence[Iteration]
        Sequence of iterations over which to compute the average.

    Returns
    -------
    ArrayLike
        The time average of `observable` over `iterations`.

    """
    for iteration in iterations:
        values = [observable(walker) for walker in iteration]
        try:
            value += np.dot(iteration.weights, values)
        except NameError:
            value = np.dot(iteration.weights, values)
    return value / len(iterations)



import logging
import math
import operator
import random
import time
from datetime import timedelta
from pickle import PickleError
from itertools import zip_longest
from collections import Counter

import numpy as np

import westpa
from .data_manager import weight_dtype
from .segment import Segment
from .states import InitialState
from . import extloader
from . import wm_ops


log = logging.getLogger(__name__)

EPS = np.finfo(weight_dtype).eps


def grouper(n, iterable, fillvalue=None):
    "Collect data into fixed-length chunks or blocks"
    # grouper(3, 'ABCDEFG', 'x') --> ABC DEF Gxx
    args = [iter(iterable)] * n
    return zip_longest(fillvalue=fillvalue, *args)


class PropagationError(RuntimeError):
    pass


class WESimManager:
    def process_config(self):
        config = self.rc.config
        for entry, type_ in [('gen_istates', bool), ('block_size', int), ('save_transition_matrices', bool)]:
            config.require_type_if_present(['west', 'propagation', entry], type_)

        self.do_gen_istates = config.get(['west', 'propagation', 'gen_istates'], False)
        self.propagator_block_size = config.get(['west', 'propagation', 'block_size'], 1)
        self.save_transition_matrices = config.get(['west', 'propagation', 'save_transition_matrices'], False)
        self.max_run_walltime = config.get(['west', 'propagation', 'max_run_wallclock'], default=None)
        self.max_total_iterations = config.get(['west', 'propagation', 'max_total_iterations'], default=None)

    def __init__(self, rc=None):
        self.rc = rc or westpa.rc
        self.work_manager = self.rc.get_work_manager()
        self.data_manager = self.rc.get_data_manager()
        self.we_driver = self.rc.get_we_driver()
        self.system = self.rc.get_system_driver()

        # A table of function -> list of (priority, name, callback) tuples
        self._callback_table = {}
        self._valid_callbacks = set(
            (
                self.prepare_run,
                self.finalize_run,
                self.prepare_iteration,
                self.finalize_iteration,
                self.pre_propagation,
                self.post_propagation,
                self.pre_we,
                self.post_we,
                self.prepare_new_iteration,
            )
        )
        self._callbacks_by_name = {fn.__name__: fn for fn in self._valid_callbacks}
        self.n_propagated = 0

        # config items
        self.do_gen_istates = False
        self.propagator_block_size = 1
        self.save_transition_matrices = False
        self.max_run_walltime = None
        self.max_total_iterations = None
        self.process_config()

        # Per-iteration variables
        self.n_iter = None  # current iteration

        # Basis and initial states for this iteration, in case the propagator needs them
        self.current_iter_bstates = None  # BasisStates valid at this iteration
        self.current_iter_istates = None  # InitialStates used in this iteration

        # Basis states for next iteration
        self.next_iter_bstates = None  # BasisStates valid for the next iteration
        self.next_iter_bstate_cprobs = None  # Cumulative probabilities for basis states, used for selection

        # Tracking of this iteration's segments
        self.segments = None  # Mapping of seg_id to segment for all segments in this iteration
        self.completed_segments = None  # Mapping of seg_id to segment for all completed segments in this iteration
        self.incomplete_segments = None  # Mapping of seg_id to segment for all incomplete segments in this iteration

        # Tracking of binning
        self.bin_mapper_hash = None  # Hash of bin mapper from most recently-run WE, for use by post-WE analysis plugins

    def register_callback(self, hook, function, priority=0):
        '''Registers a callback to execute during the given ``hook`` into the simulation loop. The optional
        priority is used to order when the function is called relative to other registered callbacks.'''

        if hook not in self._valid_callbacks:
            try:
                hook = self._callbacks_by_name[hook]
            except KeyError:
                raise KeyError('invalid hook {!r}'.format(hook))

        # It's possible to register a callback that's a duplicate function, but at a different place in memory.
        #   For example, if you launch a run without clearing state from a previous run.
        #   More details on this are available in https://github.com/westpa/westpa/issues/182 but the below code
        #   handles specifically the problem that causes in plugin loading.
        try:
            # Before checking for set membership of (priority, function.__name__, function), just check
            #   function hash for collisions in this hook.
            hook_function_hash = [hash(callback[2]) for callback in self._callback_table[hook]]
        except KeyError:
            # If there's no entry in self._callback_table for this hook, then there definitely aren't any collisions
            #   because no plugins are registered to it yet in the first place.
            pass
        else:
            # If there are plugins registered to this hook, check for duplicate hash, which will definitely have the same name, module, function.
            try:
                if hash(function) in hook_function_hash:
                    log.info('{!r} has already been loaded, skipping'.format(function))
                    return
            except KeyError:
                pass

        try:
            self._callback_table[hook].add((priority, function.__name__, function))
        except KeyError:
            self._callback_table[hook] = set([(priority, function.__name__, function)])

        # Raise warning if there are multiple callback with same priority.
        for priority, count in Counter([callback[0] for callback in self._callback_table[hook]]).items():
            if count > 1:
                log.warning(
                    f'{count} callbacks in {hook} have identical priority {priority}. The order of callback execution is not guaranteed.'
                )
                log.warning(f'{hook}: {self._callback_table[hook]}')

        log.debug('registered callback {!r} for hook {!r}'.format(function, hook))

    def invoke_callbacks(self, hook, *args, **kwargs):
        callbacks = self._callback_table.get(hook, [])
        # Sort by priority, function name, then module name
        sorted_callbacks = sorted(callbacks, key=lambda x: (x[0], x[1], x[2].__module__))
        for priority, name, fn in sorted_callbacks:
            log.debug('invoking callback {!r} for hook {!r}'.format(fn, hook))
            fn(*args, **kwargs)

    def load_plugins(self, plugins=None):
        if plugins is None:
            plugins = []

        try:
            plugins_config = westpa.rc.config['west', 'plugins']
        except KeyError:
            plugins_config = []

        if plugins_config is None:
            plugins_config = []

        plugins += plugins_config

        for plugin_config in plugins:
            plugin_name = plugin_config['plugin']
            if plugin_config.get('enabled', True):
                log.info('loading plugin {!r}'.format(plugin_name))
                plugin = extloader.get_object(plugin_name)(self, plugin_config)
                log.debug('loaded plugin {!r}'.format(plugin))

    def report_bin_statistics(self, bins, target_states, save_summary=False):
        segments = list(self.segments.values())
        bin_counts = np.fromiter(map(len, bins), dtype=np.int_, count=len(bins))
        target_counts = self.we_driver.bin_target_counts

        # Do not include bins with target count zero (e.g. sinks, never-filled bins) in the (non)empty bins statistics
        n_active_bins = len(target_counts[target_counts != 0])

        if target_states:
            n_active_bins -= len(target_states)

        seg_probs = np.fromiter(map(operator.attrgetter('weight'), segments), dtype=weight_dtype, count=len(segments))
        bin_probs = np.fromiter(map(operator.attrgetter('weight'), bins), dtype=weight_dtype, count=len(bins))
        norm = seg_probs.sum()

        assert abs(1 - norm) < EPS * (len(segments) + n_active_bins)

        min_seg_prob = seg_probs[seg_probs != 0].min()
        max_seg_prob = seg_probs.max()
        seg_drange = math.log(max_seg_prob / min_seg_prob)
        min_bin_prob = bin_probs[bin_probs != 0].min()
        max_bin_prob = bin_probs.max()
        bin_drange = math.log(max_bin_prob / min_bin_prob)
        n_pop = len(bin_counts[bin_counts != 0])

        self.rc.pstatus('{:d} of {:d} ({:%}) active bins are populated'.format(n_pop, n_active_bins, n_pop / n_active_bins))
        self.rc.pstatus('per-bin minimum non-zero probability:       {:g}'.format(min_bin_prob))
        self.rc.pstatus('per-bin maximum probability:                {:g}'.format(max_bin_prob))
        self.rc.pstatus('per-bin probability dynamic range (kT):     {:g}'.format(bin_drange))
        self.rc.pstatus('per-segment minimum non-zero probability:   {:g}'.format(min_seg_prob))
        self.rc.pstatus('per-segment maximum non-zero probability:   {:g}'.format(max_seg_prob))
        self.rc.pstatus('per-segment probability dynamic range (kT): {:g}'.format(seg_drange))
        self.rc.pstatus('norm = {:g}, error in norm = {:g} ({:.2g}*epsilon)'.format(norm, (norm - 1), (norm - 1) / EPS))
        self.rc.pflush()

        if min_seg_prob < 1e-100:
            log.warning(
                '\nMinimum segment weight is < 1e-100 and might not be physically relevant. Please reconsider your progress coordinate or binning scheme.'
            )

        if save_summary:
            iter_summary = self.data_manager.get_iter_summary()
            iter_summary['n_particles'] = len(segments)
            iter_summary['norm'] = norm
            iter_summary['min_bin_prob'] = min_bin_prob
            iter_summary['max_bin_prob'] = max_bin_prob
            iter_summary['min_seg_prob'] = min_seg_prob
            iter_summary['max_seg_prob'] = max_seg_prob
            if np.isnan(iter_summary['cputime']):
                iter_summary['cputime'] = 0.0
            if np.isnan(iter_summary['walltime']):
                iter_summary['walltime'] = 0.0
            self.data_manager.update_iter_summary(iter_summary)

    def get_bstate_pcoords(self, basis_states, label='basis'):
        '''For each of the given ``basis_states``, calculate progress coordinate values
        as necessary.  The HDF5 file is not updated.'''

        self.rc.pstatus('Calculating progress coordinate values for {} states.'.format(label))
        futures = [self.work_manager.submit(wm_ops.get_pcoord, args=(basis_state,)) for basis_state in basis_states]
        fmap = {future: i for (i, future) in enumerate(futures)}
        for future in self.work_manager.as_completed(futures):
            basis_states[fmap[future]].pcoord = future.get_result().pcoord

    def report_basis_states(self, basis_states, label='basis'):
        pstatus = self.rc.pstatus
        pstatus('{:d} {} state(s) present'.format(len(basis_states), label), end='')
        if self.rc.verbose_mode:
            pstatus(':')
            pstatus(
                '{:6s}    {:12s}    {:20s}    {:20s}    {}'.format(
                    'ID', 'Label', 'Probability', 'Aux Reference', 'Progress Coordinate'
                )
            )
            for basis_state in basis_states:
                pstatus(
                    '{:<6d}    {:12s}    {:<20.14g}    {:20s}    {}'.format(
                        basis_state.state_id,
                        basis_state.label,
                        basis_state.probability,
                        basis_state.auxref or '',
                        ', '.join(map(str, basis_state.pcoord)),
                    )
                )
        pstatus()
        self.rc.pflush()

    def report_target_states(self, target_states):
        pstatus = self.rc.pstatus
        pstatus('{:d} target state(s) present'.format(len(target_states)), end='')
        if self.rc.verbose_mode and target_states:
            pstatus(':')
            pstatus('{:6s}    {:12s}    {}'.format('ID', 'Label', 'Progress Coordinate'))
            for target_state in target_states:
                pstatus(
                    '{:<6d}    {:12s}    {}'.format(
                        target_state.state_id, target_state.label, ','.join(map(str, target_state.pcoord))
                    )
                )
        pstatus()
        self.rc.pflush()

    def initialize_simulation(self, basis_states, target_states, start_states, segs_per_state=1, suppress_we=False):
        '''Initialize a new weighted ensemble simulation, taking ``segs_per_state`` initial
        states from each of the given ``basis_states``.

        ``w_init`` is the forward-facing version of this function'''

        data_manager = self.data_manager
        work_manager = self.work_manager
        pstatus = self.rc.pstatus
        system = self.system

        pstatus('Creating HDF5 file {!r}'.format(self.data_manager.we_h5filename))
        data_manager.prepare_backing()

        # Process target states
        data_manager.save_target_states(target_states)
        self.report_target_states(target_states)

        # Process basis states
        self.get_bstate_pcoords(basis_states)
        self.data_manager.create_ibstate_group(basis_states)
        self.data_manager.create_ibstate_iter_h5file(basis_states)
        self.report_basis_states(basis_states)

        # Process start states
        # Unlike the above, does not create an ibstate group.
        # TODO: Should it? I don't think so, if needed it can be traced back through basis_auxref

        # Here, we are trying to assign a state_id to the start state to be initialized, without actually
        # saving it to the ibstates records in any of the h5 files. It might actually be a problem
        # when tracing trajectories with westpa.analysis (especially with HDF5 framework) since it would
        # try to look for a basis state id > len(basis_state).
        # Since start states are only used while initializing and iteration 1, it's ok to not save it to save space. If necessary,
        # the structure can be traced directly to the parent file using the standard basis state logic referencing
        # west[iterations/iter_00000001/ibstates/istate_index/basis_auxref] of that istate.

        if len(start_states) > 0 and start_states[0].state_id is None:
            last_id = basis_states[-1].state_id
            for start_state in start_states:
                start_state.state_id = last_id + 1
                last_id += 1

        self.get_bstate_pcoords(start_states, label='start')
        self.report_basis_states(start_states, label='start')

        pstatus('Preparing initial states')
        initial_states = []
        weights = []
        if self.do_gen_istates:
            istate_type = InitialState.ISTATE_TYPE_GENERATED
        else:
            istate_type = InitialState.ISTATE_TYPE_BASIS

        for basis_state in basis_states:
            for _iseg in range(segs_per_state):
                initial_state = data_manager.create_initial_states(1, 1)[0]
                initial_state.basis_state_id = basis_state.state_id
                initial_state.basis_state = basis_state
                initial_state.istate_type = istate_type
                weights.append(basis_state.probability / segs_per_state)
                initial_states.append(initial_state)

        for start_state in start_states:
            for _iseg in range(segs_per_state):
                initial_state = data_manager.create_initial_states(1, 1)[0]
                initial_state.basis_state_id = start_state.state_id
                initial_state.basis_state = start_state
                initial_state.basis_auxref = start_state.auxref

                # Start states are assigned their own type, so they can be identified later
                initial_state.istate_type = InitialState.ISTATE_TYPE_START
                weights.append(start_state.probability / segs_per_state)
                initial_state.iter_used = 1
                initial_states.append(initial_state)

        if self.do_gen_istates:
            futures = [
                work_manager.submit(wm_ops.gen_istate, args=(initial_state.basis_state, initial_state))
                for initial_state in initial_states
            ]
            for future in work_manager.as_completed(futures):
                rbstate, ristate = future.get_result()
                initial_states[ristate.state_id].pcoord = ristate.pcoord
        else:
            for initial_state in initial_states:
                basis_state = initial_state.basis_state
                initial_state.pcoord = basis_state.pcoord
                initial_state.istate_status = InitialState.ISTATE_STATUS_PREPARED

        for initial_state in initial_states:
            log.debug('initial state created: {!r}'.format(initial_state))

        # save list of initial states just generated
        # some of these may not be used, depending on how WE shakes out
        data_manager.update_initial_states(initial_states, n_iter=1)

        if not suppress_we:
            self.we_driver.populate_initial(initial_states, weights, system)
            segments = list(self.we_driver.next_iter_segments)
            binning = self.we_driver.next_iter_binning
        else:
            segments = list(self.we_driver.current_iter_segments)
            binning = self.we_driver.final_binning

        bin_occupancies = np.fromiter(map(len, binning), dtype=np.uint, count=self.we_driver.bin_mapper.nbins)
        target_occupancies = np.require(self.we_driver.bin_target_counts, dtype=np.uint)

        # total_bins/replicas defined here to remove target state bin from "active" bins
        total_bins = len(bin_occupancies) - len(target_states)
        total_replicas = int(sum(target_occupancies)) - int(self.we_driver.bin_target_counts[-1]) * len(target_states)

        # Make sure we have
        for segment in segments:
            segment.n_iter = 1
            segment.status = Segment.SEG_STATUS_PREPARED
            assert segment.parent_id < 0
            assert initial_states[segment.initial_state_id].iter_used == 1

        data_manager.prepare_iteration(1, segments)
        data_manager.update_initial_states(initial_states, n_iter=1)

        if self.rc.verbose_mode:
            pstatus('\nSegments generated:')
            for segment in segments:
                pstatus('{!r}'.format(segment))

        pstatus(
            '''
        Total bins:            {total_bins:d}
        Initial replicas:      {init_replicas:d} in {occ_bins:d} bins, total weight = {weight:g}
        Total target replicas: {total_replicas:d}
        '''.format(
                total_bins=total_bins,
                init_replicas=int(sum(bin_occupancies)),
                occ_bins=len(bin_occupancies[bin_occupancies > 0]),
                weight=float(sum(segment.weight for segment in segments)),
                total_replicas=total_replicas,
            )
        )

        total_prob = float(sum(segment.weight for segment in segments))
        pstatus(f'1-prob: {1 - total_prob:.4e}')

        target_counts = self.we_driver.bin_target_counts
        # Do not include bins with target count zero (e.g. sinks, never-filled bins) in the (non)empty bins statistics
        n_active_bins = len(target_counts[target_counts != 0])
        seg_probs = np.fromiter(map(operator.attrgetter('weight'), segments), dtype=weight_dtype, count=len(segments))
        norm = seg_probs.sum()

        if not abs(1 - norm) < EPS * (len(segments) + n_active_bins):
            pstatus("Normalization check failed at w_init, explicitly renormalizing")
            for segment in segments:
                segment.weight /= norm

        # Send the segments over to the data manager to commit to disk
        data_manager.current_iteration = 1

        # Report statistics
        pstatus('Simulation prepared.')
        self.segments = {segment.seg_id: segment for segment in segments}
        self.report_bin_statistics(binning, target_states, save_summary=True)
        data_manager.flush_backing()
        data_manager.close_backing()

    def prepare_iteration(self):
        log.debug('beginning iteration {:d}'.format(self.n_iter))

        # the WE driver needs a list of all target states for this iteration
        # along with information about any new weights introduced (e.g. by recycling)
        target_states = self.data_manager.get_target_states(self.n_iter)
        new_weights = self.data_manager.get_new_weight_data(self.n_iter)

        self.we_driver.new_iteration(target_states=target_states, new_weights=new_weights)

        # Get basis states used in this iteration
        self.current_iter_bstates = self.data_manager.get_basis_states(self.n_iter)

        # Get the segments for this iteration and separate into complete and incomplete
        if self.segments is None:
            segments = self.segments = {segment.seg_id: segment for segment in self.data_manager.get_segments()}
            log.debug('loaded {:d} segments'.format(len(segments)))
        else:
            segments = self.segments
            log.debug('using {:d} pre-existing segments'.format(len(segments)))

        completed_segments = self.completed_segments = {}
        incomplete_segments = self.incomplete_segments = {}
        for segment in segments.values():
            if segment.status == Segment.SEG_STATUS_COMPLETE:
                completed_segments[segment.seg_id] = segment
            else:
                incomplete_segments[segment.seg_id] = segment
        log.debug('{:d} segments are complete; {:d} are incomplete'.format(len(completed_segments), len(incomplete_segments)))

        if len(incomplete_segments) == len(segments):
            # Starting a new iteration
            self.rc.pstatus('Beginning iteration {:d}'.format(self.n_iter))
        elif incomplete_segments:
            self.rc.pstatus('Continuing iteration {:d}'.format(self.n_iter))
        self.rc.pstatus(
            '{:d} segments remain in iteration {:d} ({:d} total)'.format(len(incomplete_segments), self.n_iter, len(segments))
        )

        # Get the initial states active for this iteration (so that the propagator has them if necessary)
        self.current_iter_istates = {
            state.state_id: state for state in self.data_manager.get_segment_initial_states(list(segments.values()))
        }
        log.debug('This iteration uses {:d} initial states'.format(len(self.current_iter_istates)))

        # Assign this iteration's segments' initial points to bins and report on bin population
        initial_pcoords = self.system.new_pcoord_array(len(segments))
        initial_binning = self.system.bin_mapper.construct_bins()
        for iseg, segment in enumerate(segments.values()):
            initial_pcoords[iseg] = segment.pcoord[0]
        initial_assignments = self.system.bin_mapper.assign(initial_pcoords)
        for segment, assignment in zip(iter(segments.values()), initial_assignments):
            initial_binning[assignment].add(segment)
        self.report_bin_statistics(initial_binning, [], save_summary=True)
        del initial_pcoords, initial_binning

        self.rc.pstatus('Waiting for segments to complete...')

        # Let the WE driver assign completed segments
        if completed_segments:
            self.we_driver.assign(list(completed_segments.values()))

        # load restart data
        self.data_manager.prepare_segment_restarts(
            incomplete_segments.values(), self.current_iter_bstates, self.current_iter_istates
        )

        # Get the basis states and initial states for the next iteration, necessary for doing on-the-fly recycling
        self.next_iter_bstates = self.data_manager.get_basis_states(self.n_iter + 1)
        self.next_iter_bstate_cprobs = np.add.accumulate([bstate.probability for bstate in self.next_iter_bstates])

        self.we_driver.avail_initial_states = {
            istate.state_id: istate for istate in self.data_manager.get_unused_initial_states(n_iter=self.n_iter + 1)
        }
        log.debug('{:d} unused initial states found'.format(len(self.we_driver.avail_initial_states)))

        # Invoke callbacks
        self.invoke_callbacks(self.prepare_iteration)

        # dispatch and immediately wait on result for prep_iter
        log.debug('dispatching propagator prep_iter to work manager')
        self.work_manager.submit(wm_ops.prep_iter, args=(self.n_iter, segments)).get_result()

    def finalize_iteration(self):
        '''Clean up after an iteration and prepare for the next.'''
        log.debug('finalizing iteration {:d}'.format(self.n_iter))

        self.invoke_callbacks(self.finalize_iteration)

        # dispatch and immediately wait on result for post_iter
        log.debug('dispatching propagator post_iter to work manager')
        self.work_manager.submit(wm_ops.post_iter, args=(self.n_iter, list(self.segments.values()))).get_result()

        # Move existing segments into place as new segments
        del self.segments
        self.segments = {segment.seg_id: segment for segment in self.we_driver.next_iter_segments}

        self.rc.pstatus("Iteration completed successfully")

    def get_istate_futures(self):
        '''Add ``n_states`` initial states to the internal list of initial states assigned to
        recycled particles.  Spare states are used if available, otherwise new states are created.
        If created new initial states requires generation, then a set of futures is returned
        representing work manager tasks corresponding to the necessary generation work.'''

        n_recycled = self.we_driver.n_recycled_segs
        n_istates_needed = self.we_driver.n_istates_needed

        log.debug('{:d} unused initial states available'.format(len(self.we_driver.avail_initial_states)))
        log.debug('{:d} new initial states required for recycling {:d} walkers'.format(n_istates_needed, n_recycled))

        futures = set()
        updated_states = []
        for _i in range(n_istates_needed):
            # Select a basis state according to its weight
            ibstate = np.digitize([random.random()], self.next_iter_bstate_cprobs)
            basis_state = self.next_iter_bstates[ibstate[0]]
            initial_state = self.data_manager.create_initial_states(1, n_iter=self.n_iter + 1)[0]
            initial_state.iter_created = self.n_iter
            initial_state.basis_state_id = basis_state.state_id
            initial_state.istate_status = InitialState.ISTATE_STATUS_PENDING

            if self.do_gen_istates:
                log.debug('generating new initial state from basis state {!r}'.format(basis_state))
                initial_state.istate_type = InitialState.ISTATE_TYPE_GENERATED
                futures.add(self.work_manager.submit(wm_ops.gen_istate, args=(basis_state, initial_state)))
            else:
                log.debug('using basis state {!r} directly'.format(basis_state))
                initial_state.istate_type = InitialState.ISTATE_TYPE_BASIS
                initial_state.pcoord = basis_state.pcoord.copy()
                initial_state.istate_status = InitialState.ISTATE_STATUS_PREPARED
                self.we_driver.avail_initial_states[initial_state.state_id] = initial_state
            updated_states.append(initial_state)
        self.data_manager.update_initial_states(updated_states, n_iter=self.n_iter + 1)
        return futures

    def propagate(self):
        segments = list(self.incomplete_segments.values())
        log.debug('iteration {:d}: propagating {:d} segments'.format(self.n_iter, len(segments)))

        # all futures dispatched for this iteration
        futures = set()
        segment_futures = set()

        # Immediately dispatch any necessary initial state generation
        istate_gen_futures = self.get_istate_futures()
        futures.update(istate_gen_futures)

        # Dispatch propagation tasks using work manager
        for segment_block in grouper(self.propagator_block_size, segments):
            segment_block = [_f for _f in segment_block if _f]
            pbstates, pistates = westpa.core.states.pare_basis_initial_states(
                self.current_iter_bstates, list(self.current_iter_istates.values()), segment_block
            )
            future = self.work_manager.submit(wm_ops.propagate, args=(pbstates, pistates, segment_block))
            futures.add(future)
            segment_futures.add(future)

        while futures:
            # TODO: add capacity for timeout or SIGINT here
            future = self.work_manager.wait_any(futures)
            futures.remove(future)

            if future in segment_futures:
                segment_futures.remove(future)
                incoming = future.get_result()
                self.n_propagated += 1

                self.segments.update({segment.seg_id: segment for segment in incoming})
                self.completed_segments.update({segment.seg_id: segment for segment in incoming})

                self.we_driver.assign(incoming)
                new_istate_futures = self.get_istate_futures()
                istate_gen_futures.update(new_istate_futures)
                futures.update(new_istate_futures)

                with self.data_manager.expiring_flushing_lock():
                    self.data_manager.update_segments(self.n_iter, incoming)

            elif future in istate_gen_futures:
                istate_gen_futures.remove(future)
                _basis_state, initial_state = future.get_result()
                log.debug('received newly-prepared initial state {!r}'.format(initial_state))
                initial_state.istate_status = InitialState.ISTATE_STATUS_PREPARED
                with self.data_manager.expiring_flushing_lock():
                    self.data_manager.update_initial_states([initial_state], n_iter=self.n_iter + 1)
                self.we_driver.avail_initial_states[initial_state.state_id] = initial_state
            else:
                log.error('unknown future {!r} received from work manager'.format(future))
                raise AssertionError('untracked future {!r}'.format(future))

        log.debug('done with propagation')
        self.save_bin_data()
        self.data_manager.flush_backing()

    def save_bin_data(self):
        '''Calculate and write flux and transition count matrices to HDF5. Population and rate matrices
        are likely useless at the single-tau level and are no longer written.'''
        # save_bin_data(self, populations, n_trans, fluxes, rates, n_iter=None)

        if self.save_transition_matrices:
            with self.data_manager.expiring_flushing_lock():
                iter_group = self.data_manager.get_iter_group(self.n_iter)
                for key in ['bin_ntrans', 'bin_fluxes']:
                    try:
                        del iter_group[key]
                    except KeyError:
                        pass
                iter_group['bin_ntrans'] = self.we_driver.transition_matrix
                iter_group['bin_fluxes'] = self.we_driver.flux_matrix

    def check_propagation(self):
        '''Check for failures in propagation or initial state generation, and raise an exception
        if any are found.'''

        failed_segments = [segment for segment in self.segments.values() if segment.status != Segment.SEG_STATUS_COMPLETE]

        if failed_segments:
            failed_ids = '  \n'.join(str(segment.seg_id) for segment in failed_segments)
            log.error('propagation failed for {:d} segment(s):\n{}'.format(len(failed_segments), failed_ids))
            raise PropagationError('propagation failed for {:d} segments'.format(len(failed_segments)))
        else:
            log.debug('propagation complete for iteration {:d}'.format(self.n_iter))

        failed_istates = [
            istate
            for istate in self.we_driver.used_initial_states.values()
            if istate.istate_status != InitialState.ISTATE_STATUS_PREPARED
        ]
        log.debug('{!r}'.format(failed_istates))
        if failed_istates:
            failed_ids = '  \n'.join(str(istate.state_id) for istate in failed_istates)
            log.error('initial state generation failed for {:d} states:\n{}'.format(len(failed_istates), failed_ids))
            raise PropagationError('initial state generation failed for {:d} states'.format(len(failed_istates)))
        else:
            log.debug('initial state generation complete for iteration {:d}'.format(self.n_iter))

    def run_we(self):
        '''Run the weighted ensemble algorithm based on the binning in self.final_bins and
        the recycled particles in self.to_recycle, creating and committing the next iteration's
        segments to storage as well.'''

        # The WE driver now does almost everything; we just have to record the
        # mapper used for binning this iteration, and update initial states
        # that have been used

        try:
            pickled, hashed = self.we_driver.bin_mapper.pickle_and_hash()
        except PickleError:
            pickled = hashed = ''

        self.bin_mapper_hash = hashed
        self.we_driver.construct_next()

        if self.we_driver.used_initial_states:
            for initial_state in self.we_driver.used_initial_states.values():
                initial_state.iter_used = self.n_iter + 1
            self.data_manager.update_initial_states(list(self.we_driver.used_initial_states.values()))

        self.data_manager.update_segments(self.n_iter, list(self.segments.values()))

        self.data_manager.require_iter_group(self.n_iter + 1)
        self.data_manager.save_iter_binning(self.n_iter + 1, hashed, pickled, self.we_driver.bin_target_counts)

        # Report on recycling
        recycling_events = {}
        for nw in self.we_driver.new_weights:
            try:
                recycling_events[nw.target_state_id].append(nw.weight)
            except KeyError:
                recycling_events[nw.target_state_id] = list([nw.weight])

        tstates_by_id = {state.state_id: state for state in self.we_driver.target_states.values()}

        for tstate_id, weights in recycling_events.items():
            tstate = tstates_by_id[tstate_id]
            self.rc.pstatus(
                'Recycled {:g} probability ({:d} walkers) from target state {!r}'.format(sum(weights), len(weights), tstate.label)
            )

    def prepare_new_iteration(self):
        '''Commit data for the coming iteration to the HDF5 file.'''
        self.invoke_callbacks(self.prepare_new_iteration)

        if self.rc.debug_mode:
            self.rc.pstatus('\nSegments generated:')
            for segment in self.we_driver.next_iter_segments:
                self.rc.pstatus('{!r} pcoord[0]={!r}'.format(segment, segment.pcoord[0]))

        self.data_manager.prepare_iteration(self.n_iter + 1, list(self.we_driver.next_iter_segments))
        self.data_manager.save_new_weight_data(self.n_iter + 1, self.we_driver.new_weights)

    def run(self):
        run_starttime = time.time()
        max_walltime = self.max_run_walltime
        if max_walltime:
            run_killtime = run_starttime + max_walltime
            self.rc.pstatus('Maximum wallclock time: %s' % timedelta(seconds=max_walltime or 0))
        else:
            run_killtime = None

        self.n_iter = self.data_manager.current_iteration
        max_iter = self.max_total_iterations or self.n_iter + 1

        iter_elapsed = 0
        while self.n_iter <= max_iter:
            if max_walltime and time.time() + 1.1 * iter_elapsed >= run_killtime:
                self.rc.pstatus('Iteration {:d} would require more than the allotted time. Ending run.'.format(self.n_iter))
                return

            try:
                iter_start_time = time.time()

                self.rc.pstatus('\n%s' % time.asctime())
                self.rc.pstatus('Iteration %d (%d requested)' % (self.n_iter, max_iter))

                self.prepare_iteration()
                self.rc.pflush()

                self.pre_propagation()
                self.propagate()
                self.rc.pflush()
                self.check_propagation()
                self.rc.pflush()
                self.post_propagation()

                cputime = sum(segment.cputime for segment in self.segments.values())

                self.rc.pflush()
                self.pre_we()
                self.run_we()
                self.post_we()
                self.rc.pflush()

                self.prepare_new_iteration()

                self.finalize_iteration()

                iter_elapsed = time.time() - iter_start_time
                iter_summary = self.data_manager.get_iter_summary()
                iter_summary['walltime'] += iter_elapsed
                iter_summary['cputime'] = cputime
                self.data_manager.update_iter_summary(iter_summary)

                self.n_iter += 1
                self.data_manager.current_iteration += 1

                try:
                    # This may give NaN if starting a truncated simulation
                    walltime = timedelta(seconds=float(iter_summary['walltime']))
                except ValueError:
                    walltime = 0.0

                try:
                    cputime = timedelta(seconds=float(iter_summary['cputime']))
                except ValueError:
                    cputime = 0.0

                self.rc.pstatus('Iteration wallclock: {0!s}, cputime: {1!s}\n'.format(walltime, cputime))
                self.rc.pflush()
            finally:
                self.data_manager.flush_backing()

        self.rc.pstatus('\n%s' % time.asctime())
        self.rc.pstatus('WEST run complete.')

    def prepare_run(self):
        '''Prepare a new run.'''
        self.data_manager.prepare_run()
        self.system.prepare_run()
        self.invoke_callbacks(self.prepare_run)

    def finalize_run(self):
        '''Perform cleanup at the normal end of a run'''
        self.invoke_callbacks(self.finalize_run)
        self.system.finalize_run()
        self.data_manager.finalize_run()

    def pre_propagation(self):
        self.invoke_callbacks(self.pre_propagation)

    def post_propagation(self):
        self.invoke_callbacks(self.post_propagation)

    def pre_we(self):
        self.invoke_callbacks(self.pre_we)

    def post_we(self):
        self.invoke_callbacks(self.post_we)



import argparse
import io
import logging
import sys

import numpy as np

import westpa.work_managers as work_managers
from westpa.work_managers import make_work_manager

import westpa
from westpa.core.segment import Segment
from westpa.core.states import BasisState, TargetState


log = logging.getLogger('w_states')
EPS = np.finfo(np.float64).eps


def entry_point():
    parser = argparse.ArgumentParser(
        'w_states',
        description='''\
    Display or manipulate basis (initial) or target (recycling) states for a WEST simulation.  By default, states are
    displayed (or dumped to files).  If ``--replace`` is specified, all basis/target states are replaced for the
    next iteration.  If ``--append`` is specified, the given target state(s) are appended to the list for the
    next iteration.

    Appending basis states is not permitted, as this would require renormalizing basis state
    probabilities in ways that may be error-prone. Instead, use ``w_states --show --bstate-file=bstates.txt``
    and then edit the resulting ``bstates.txt`` file to include the new desired basis states, then use
    ``w_states --replace --bstate-file=bstates.txt`` to update the WEST HDF5 file appropriately.
    ''',
    )
    westpa.rc.add_args(parser)
    smgroup = parser.add_argument_group('modes of operation')
    mode_group = smgroup.add_mutually_exclusive_group()
    mode_group.add_argument(
        '--show', dest='mode', action='store_const', const='show', help='Display current basis/target states (or dump to files).'
    )
    mode_group.add_argument(
        '--append',
        dest='mode',
        action='store_const',
        const='append',
        help='Append the given basis/target states to those currently in use.',
    )
    mode_group.add_argument(
        '--replace',
        dest='mode',
        action='store_const',
        const='replace',
        help='Replace current basis/target states with those specified.',
    )
    parser.add_argument(
        '--bstate-file',
        metavar='BSTATE_FILE',
        help='''Read (--append/--replace) or write (--show) basis state names, probabilities,
                        and data references from/to BSTATE_FILE.''',
    )
    parser.add_argument(
        '--bstate',
        action='append',
        dest='bstates',
        help='''Add the given basis state (specified as a string 'label,probability[,auxref]')
                        to the list of basis states (after those specified in --bstate-file, if any). This argument
                        may be specified more than once, in which case the given states are appended in the order
                        they are given on the command line.''',
    )
    parser.add_argument(
        '--tstate-file',
        metavar='TSTATE_FILE',
        help='''Read (--append/--replace) or write (--show) target state names
                        and representative progress coordinates from/to TSTATE_FILE''',
    )
    parser.add_argument(
        '--tstate',
        action='append',
        dest='tstates',
        help='''Add the given target state (specified as a string 'label,pcoord0[,pcoord1[,...]]') to the
                        list of target states (after those specified in the file given by --tstates-from, if any).
                        This argument may be specified more than once, in which case the given states are appended
                        in the order they appear on the command line.''',
    )
    parser.set_defaults(mode='show')

    work_managers.environment.add_wm_args(parser)
    args = parser.parse_args()
    westpa.rc.process_args(args)
    work_managers.environment.process_wm_args(args)

    # Need to have something to pass to initialize
    if not hasattr(args, 'bstates'):
        args.bstates = None
    if not hasattr(args, 'tstates'):
        args.tstates = None
    if not hasattr(args, 'tstate_file'):
        args.tstate_file = None

    initialize(args.mode, args.bstates, args.bstate_file, args.tstates, args.tstate_file)


# TODO: This would benefit from a refactor to set default args to None, and replace some of those "if <argument>" clauses
def initialize(mode, bstates, _bstate_file, tstates, _tstate_file):
    work_manager = make_work_manager()

    system = westpa.rc.get_system_driver()

    with work_manager:
        if work_manager.is_master:
            data_manager = westpa.rc.get_data_manager()
            data_manager.open_backing(mode='a')
            sim_manager = westpa.rc.get_sim_manager()
            n_iter = data_manager.current_iteration

            assert mode in ('show', 'replace', 'append')
            if mode == 'show':
                basis_states = data_manager.get_basis_states(n_iter)
                if basis_states:
                    bstate_file = sys.stdout if not _bstate_file else open(_bstate_file, 'wt')
                    bstate_file.write('# Basis states for iteration {:d}\n'.format(n_iter))
                    BasisState.states_to_file(basis_states, bstate_file)

                target_states = data_manager.get_target_states(n_iter)
                if target_states:
                    tstate_file = sys.stdout if not _tstate_file else open(_tstate_file, 'wt')
                    tstate_file.write('# Target states for iteration {:d}\n'.format(n_iter))
                    TargetState.states_to_file(target_states, tstate_file)

            elif mode == 'replace':
                seg_index = data_manager.get_seg_index(n_iter)
                if (seg_index['status'] == Segment.SEG_STATUS_COMPLETE).any():
                    print('Iteration {:d} has completed segments; applying new states to iteration {:d}'.format(n_iter, n_iter + 1))
                    n_iter += 1

                basis_states = []
                if _bstate_file:
                    basis_states.extend(BasisState.states_from_file(_bstate_file))
                if bstates:
                    for bstate_str in bstates:
                        fields = bstate_str.split(',')
                        label = fields[0]
                        probability = float(fields[1])
                        try:
                            auxref = fields[2]
                        except IndexError:
                            auxref = None
                        basis_states.append(BasisState(label=label, probability=probability, auxref=auxref))

                if basis_states:
                    # Check that the total probability of basis states adds to one
                    tprob = sum(bstate.probability for bstate in basis_states)
                    if abs(1.0 - tprob) > len(basis_states) * EPS:
                        pscale = 1 / tprob
                        log.warning('Basis state probabilities do not add to unity; rescaling by {:g}'.format(pscale))
                        for bstate in basis_states:
                            bstate.probability *= pscale

                    # Assign progress coordinates to basis states
                    sim_manager.get_bstate_pcoords(basis_states)
                    data_manager.create_ibstate_group(basis_states, n_iter)
                    sim_manager.report_basis_states(basis_states)

                # Now handle target states
                target_states = []
                if _tstate_file:
                    target_states.extend(TargetState.states_from_file(_tstate_file, system.pcoord_dtype))
                if tstates:
                    tstates_strio = io.StringIO('\n'.join(tstates).replace(',', ' '))
                    target_states.extend(TargetState.states_from_file(tstates_strio, system.pcoord_dtype))
                    del tstates_strio

                if not target_states:
                    westpa.rc.pstatus('No target states specified.')
                else:
                    data_manager.save_target_states(target_states, n_iter)
                    sim_manager.report_target_states(target_states)

                data_manager.update_iter_group_links(n_iter)

            else:  # args.mode == 'append'
                if _bstate_file or bstates:
                    sys.stderr.write('refusing to append basis states; use --show followed by --replace instead\n')
                    sys.exit(2)

                target_states = data_manager.get_target_states(n_iter)

                seg_index = data_manager.get_seg_index(n_iter)
                if (seg_index['status'] == Segment.SEG_STATUS_COMPLETE).any():
                    print('Iteration {:d} has completed segments; applying new states to iteration {:d}'.format(n_iter, n_iter + 1))
                    n_iter += 1

                if _tstate_file:
                    target_states.extend(TargetState.states_from_file(_tstate_file, system.pcoord_dtype))
                if tstates:
                    tstates_strio = io.StringIO('\n'.join(tstates).replace(',', ' '))
                    target_states.extend(TargetState.states_from_file(tstates_strio, system.pcoord_dtype))
                    del tstates_strio

                if not target_states:
                    westpa.rc.pstatus('No target states specified.')
                else:
                    data_manager.save_target_states(target_states, n_iter)
                    sim_manager.report_target_states(target_states)

                data_manager.update_iter_group_links(n_iter)
        else:
            work_manager.run()


if __name__ == '__main__':
    entry_point()




from __future__ import division
import numpy, sys, math, os
cimport numpy, cython
from cpython.buffer cimport *
from numpy cimport *

from libc.math cimport floor, ceil, log10
#from libc.stdlib cimport RAND_MAX, rand, srand

ctypedef fused _fptype:
    numpy.float32_t
    numpy.float64_t

# Number of bytes in a dataset above which correlation time estimates are
# not performed by generating synthetic data sets once and then evaluating
# the autocorrelation elements on them, but rather by generating a new
# synthetic dataset each time an autocorrelation element is required.
# Essentially, this limits the RAM used during the correlation time
# calculation to approximately CORRELTIME_CROSSOVER.  Keep in mind that
# the space savings is at the cost of needing *many* more random numbers,
# which will almost certainly become the rate-limiting step of the
# calculation.
cdef Py_ssize_t CORRELTIME_CROSSOVER = 512*1024*1024


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef _fptype _autocorrel_elem(_fptype[:] xs, long k):
    cdef:
        long N = xs.shape[0]
        long i
        double cdif
        double xbar = 0.0
        double norm = 0.0
        double rho = 0.0

    with nogil:
        # calculate the mean manually, because it's faster than calling out to
        # numpy.mean
        for i in range(N):
            xbar += xs[i]
        xbar /= N

        for i in range(N):
            cdif = xs[i] - xbar
            norm += cdif*cdif

            if i < N-k:
                rho += cdif * (xs[i+k]-xbar)

        rho *= N / ((N-k)*norm)

    return <_fptype> rho

cpdef autocorrel_elem(xs, k):
    '''Calculate the ``k``-lag sample autocorrelation of ``xs``. ``xs`` must be float32 or float64.'''

    cdef:
        int typecode = PyArray_TYPE(xs)

    if typecode == NPY_FLOAT64:
        return _autocorrel_elem[numpy.float64_t](xs,k)
    elif typecode == NPY_FLOAT32:
        return _autocorrel_elem[numpy.float32_t](xs,k)
    else:
        raise TypeError('unsupported type')

cpdef Py_ssize_t get_bssize(double alpha) nogil:
    '''Return a bootstrap data set size appropriate for the given confidence level.'''

    cdef:
        Py_ssize_t bssize=1, i

    #return int(10**(math.ceil(-math.log10(alpha)) + 1))
    for i in range(<long> ceil(-log10(alpha)) + 1):
        bssize *= 10
    return bssize

cpdef mcbs_ci(dataset, estimator, alpha, dlen, n_sets=None, args=None, kwargs=None, sort=numpy.msort):
    '''Perform a Monte Carlo bootstrap estimate for the (1-``alpha``) confidence interval
    on the given ``dataset`` with the given ``estimator``.  This routine is not appropriate
    for time-correlated data.

    Returns ``(estimate, ci_lb, ci_ub)`` where ``estimate`` is the application of the
    given ``estimator`` to the input ``dataset``, and ``ci_lb`` and ``ci_ub`` are the
    lower and upper limits, respectively, of the (1-``alpha``) confidence interval on
    ``estimate``.

    ``estimator`` is called as ``estimator(dataset, *args, **kwargs)``. Common estimators include:
      * numpy.mean -- calculate the confidence interval on the mean of ``dataset``
      * numpy.median -- calculate a confidence interval on the median of ``dataset``
      * numpy.std -- calculate a confidence interval on the standard deviation of ``datset``.

    ``n_sets`` is the number of synthetic data sets to generate using the given ``estimator``,
    which will be chosen using `get_bssize()`_ if ``n_sets`` is not given.

    ``sort`` can be used
    to override the sorting routine used to calculate the confidence interval, which should
    only be necessary for estimators returning vectors rather than scalars.
    '''

    if alpha > 0.5:
        raise ValueError('alpha ({}) > 0.5'.format(alpha))

    args = args or ()
    kwargs = kwargs or {}

    # dataset SHOULD be a dictionary.
    d_input = dataset.copy()
    # Here, we're dumping in any extra kwarg arguments to pass in to the estimator.
    try:
        d_input.update(kwargs)
    except Exception:
        pass

    fhat = estimator(**d_input)

    try:
        estimator_shape = fhat.shape
    except AttributeError:
        estimator_shape = ()

    try:
        estimator_dtype = fhat.dtype
    except AttributeError:
        estimator_dtype = type(fhat)

    n_sets = n_sets or get_bssize(alpha)

    f_synth = numpy.empty((n_sets,) + estimator_shape, dtype=estimator_dtype)

    for i in xrange(n_sets):
        indices = numpy.random.randint(dlen, size=(dlen,))
        d_synth = {}
        for key, dset in dataset.iteritems():
            d_synth[key] = numpy.take(dset, indices, axis=0)
        d_input = d_synth.copy()
        try:
            d_input.update(kwargs)
        except Exception:
            pass
        f_synth[i] = estimator(**d_input)
        del indices

    f_synth_sorted = sort(f_synth)
    lbi = int(math.floor(n_sets*alpha/2.0))
    ubi = int(math.ceil(n_sets*(1-alpha/2.0)))
    lb = f_synth_sorted[lbi]
    ub = f_synth_sorted[ubi]
    sterr = numpy.std(f_synth_sorted)

    del f_synth_sorted, f_synth
    return (fhat, lb, ub, sterr)

cpdef mcbs_correltime(dataset, alpha, n_sets = None):
    '''Calculate the correlation time of the given ``dataset``, significant to the
    (1-``alpha``) level, using the method described in Huber & Kim, "Weighted-ensemble
    Brownian dynamics simulations for protein association reactions" (1996),
    doi:10.1016/S0006-3495(96)79552-8. An appropriate balance between space and speed
    is chosen based on the size of the input data.

    Returns 0 for data statistically uncorrelated with (1-alpha) confidence, otherwise
    the correlation length. (Thus, the appropriate stride for blocking is the
    result of this function plus one.)'''

    n_sets = n_sets or get_bssize(alpha)
    if dataset.nbytes * n_sets > CORRELTIME_CROSSOVER:
        return mcbs_correltime_fullauto(dataset, alpha, n_sets)
    else:
        return mcbs_correltime_small(dataset, alpha, n_sets)

cpdef mcbs_correltime_fullauto(dataset, alpha, n_sets):
    '''Specialization of `mcbs_correltime`_ for large datasets, where rather than generating
    ``n_sets`` synthetic datasets and then evaluating autocorrelation elements on them, a
    synthetic data set is generated for each autocorrelation element required. This trades time
    for space.'''

    for k in xrange(1,len(dataset)//2):
        data_rho, synrho_lb, synrho_ub = mcbs_ci(dataset, autocorrel_elem, alpha, n_sets, args=(k,))
        if data_rho > synrho_lb and data_rho < synrho_ub:
            return k-1
    else:
        return len(dataset)
        #raise ValueError('correlation length exceeds half the data set length')

cpdef mcbs_correltime_small(dataset, alpha, n_sets):
    '''Specialization of `mcbs_correltime`_ for small-to-mediam datasets, where ``n_sets``
    synthetic datasets are generated and stored, and then autocorrelation elements
    calculated on them. This implementation trades space for time.'''

    if alpha > 0.5:
        raise ValueError('alpha ({}) > 0.5'.format(alpha))

    dlen = len(dataset)
    lbi = int(math.floor(n_sets*alpha/2.0))
    ubi = int(math.ceil(n_sets*(1-alpha/2.0)))

    synth_sets = numpy.empty((n_sets,)+dataset.shape, dtype=dataset.dtype)
    synth_acf_elems = numpy.empty((n_sets,), numpy.float64)

    for i in xrange(n_sets):
        synth_sets[i] = numpy.take(dataset,numpy.random.randint(dlen,size=(dlen,)))

    for k in xrange(1,dlen//2):
        data_rho = autocorrel_elem(dataset, k)

        for i in xrange(n_sets):
            synth_acf_elems[i] = autocorrel_elem(synth_sets[i],k)

        synth_acf_elems.sort()

        if data_rho > synth_acf_elems[lbi] and data_rho < synth_acf_elems[ubi]:
            return k-1
    else:
        #raise ValueError('correlation length exceeds half the data set length')
        return len(dataset)





'''A package for performing Monte Carlo bootstrap estimates of
statistics.'''

import numpy as np

from ._mclib import mcbs_correltime, get_bssize, mcbs_ci


def mcbs_ci_correl(
    estimator_datasets,
    estimator,
    alpha,
    n_sets=None,
    args=None,
    autocorrel_alpha=None,
    autocorrel_n_sets=None,
    subsample=None,
    do_correl=True,
    mcbs_enable=None,
    estimator_kwargs={},
):
    '''Perform a Monte Carlo bootstrap estimate for the (1-``alpha``) confidence interval
    on the given ``dataset`` with the given ``estimator``.  This routine is appropriate
    for time-correlated data, using the method described in Huber & Kim, "Weighted-ensemble
    Brownian dynamics simulations for protein association reactions" (1996),
    doi:10.1016/S0006-3495(96)79552-8 to determine a statistically-significant correlation time
    and then reducing the dataset by a factor of that correlation time before running a "classic"
    Monte Carlo bootstrap.

    Returns ``(estimate, ci_lb, ci_ub, correl_time)`` where ``estimate`` is the application of the
    given ``estimator`` to the input ``dataset``, ``ci_lb`` and ``ci_ub`` are the
    lower and upper limits, respectively, of the (1-``alpha``) confidence interval on
    ``estimate``, and ``correl_time`` is the correlation time of the dataset, significant to
    (1-``autocorrel_alpha``).

    ``estimator`` is called as ``estimator(dataset, *args, **kwargs)``. Common estimators include:
      * np.mean -- calculate the confidence interval on the mean of ``dataset``
      * np.median -- calculate a confidence interval on the median of ``dataset``
      * np.std -- calculate a confidence interval on the standard deviation of ``datset``.

    ``n_sets`` is the number of synthetic data sets to generate using the given ``estimator``,
    which will be chosen using `get_bssize()`_ if ``n_sets`` is not given.

    ``autocorrel_alpha`` (which defaults to ``alpha``) can be used to adjust the significance
    level of the autocorrelation calculation. Note that too high a significance level (too low an
    alpha) for evaluating the significance of autocorrelation values can result in a failure to
    detect correlation if the autocorrelation function is noisy.

    The given ``subsample`` function is used, if provided, to subsample the dataset prior to running
    the full Monte Carlo bootstrap. If none is provided, then a random entry from each correlated
    block is used as the value for that block.  Other reasonable choices include ``np.mean``,
    ``np.median``, ``(lambda x: x[0])`` or ``(lambda x: x[-1])``.  In particular, using
    ``subsample=np.mean`` will converge to the block averaged mean and standard error,
    while accounting for any non-normality in the distribution of the mean.
    '''

    if alpha > 0.5:
        raise ValueError('alpha ({}) > 0.5'.format(alpha))

    autocorrel_alpha = alpha if not autocorrel_alpha else autocorrel_alpha

    # We're now passing in dataset as a dict, so we need to enforce that for compatibility with older tools.
    # This just takes our dataset and puts it into a dict, as it's likely that we're using
    # mean or median as our estimators, which take "a" as argument input.
    if not isinstance(estimator_datasets, dict):
        # Enforcing the data structure.
        pre_calculated = estimator_datasets
        estimator_datasets = {'a': estimator_datasets}
        # This also probably means our estimator isn't going to handle kwargs, so we'll watch out for that later in testing.
        # We may have to replace the 'simple' estimator with a slightly more complex lambda function which simply ditches extra arguments.
    for key, dset in estimator_datasets.items():
        estimator_datasets[key] = np.asanyarray(dset)
        dlen = dset.shape[0]

    # Why do we have 'estimator_datasets'?
    # Estimators may require many different sets of data to properly function; while we can send this in via the kwargs,
    # we may wish to decimate only a certain subset (due to the block bootstrapping) of the input parameters.
    # Therefore, 'estimator_datasets' should consist of datasets that must be sliced/decimated with the subsampling function.
    # Some estimators (such as the reweighting) may not be able to be decimated in a straightforward manner with a subsample function,
    # as we cannot pre-estimate the quantity without introducing error or bias.  In those cases, we may wish to pass on all the data,
    # but ensure that our estimator only includes certain iterations (and only in a certain way).

    n_sets = n_sets or get_bssize(alpha)
    autocorrel_n_sets = autocorrel_n_sets or get_bssize(autocorrel_alpha)

    if mcbs_enable is False:
        # While it's odd to support NOT doing the bootstrap in a library specifically designed for bootstrapping,
        # supporting this functionality here makes writing the code a lot easier, as we can just pass in a flag.
        # Specifically, this is for situations in which error is not desired (that is, only a reasonable mean is desired).
        # It's often useful when doing a quick analysis.
        estimator_datasets.update(estimator_kwargs)
        try:
            estimator_datasets.update({'stride': 1})
        except Exception:
            pass

        return_set = estimator(**estimator_datasets)
        # We don't try and pretend we're doing any error analysis.
        return return_set, return_set, return_set, 0, 1

    # We need to pre-generate the data; why not do it here?  We're already set up for it...
    precalc_kwargs = estimator_kwargs.copy()
    precalc_kwargs['stride'] = 1
    pre_calculated = []
    for block in range(1, dlen + 1):
        for key, dset in estimator_datasets.items():
            precalc_kwargs[key] = dset[0:block]
        pre_calculated.append(estimator(**precalc_kwargs))
    # We need to get rid of any NaNs.
    pre_calculated = np.asanyarray(pre_calculated)
    pre_calculated = pre_calculated[np.isfinite(pre_calculated)]
    # If this happens, we have a huge NaN problem.  That is, our estimator is failing to return meaningful
    # numbers.  We should catch this when it happens, and so raise an exception, here.
    # This is almost certainly due to estimator failure.  Double check that calculation.
    if pre_calculated.shape == (0,):
        raise NameError("Looks like the estimator failed.  This is likely a programming issue, and should be reported.")
    # If pre-calculated is not None, we'll use that instead of dataset.
    # We can also assume that it's a 1 dimensional set with nothing needed, so 'key' should work.
    if do_correl is True:
        correl_len = mcbs_correltime(pre_calculated, autocorrel_alpha, autocorrel_n_sets)
    else:
        correl_len = 0
    if correl_len == len(pre_calculated):
        # too correlated for meaningful calculations
        estimator_datasets.update(estimator_kwargs)
        try:
            estimator_datasets.update({'stride': 1})
        except Exception:
            pass

        return estimator(**estimator_datasets), pre_calculated.min(), pre_calculated.max(), (np.std(pre_calculated)), correl_len

    # else, do a blocked bootstrap
    stride = correl_len + 1

    if stride == 1:
        # Some estimators may require the stride, so we pass it in.
        estimator_kwargs['stride'] = stride
        return mcbs_ci(
            dataset=estimator_datasets,
            estimator=estimator,
            alpha=alpha,
            dlen=dlen,
            n_sets=n_sets,
            args=args,
            kwargs=estimator_kwargs,
            sort=np.msort,
        ) + (correl_len,)
    else:
        subsample = subsample or (lambda x: x[np.random.randint(len(x))])
        # Let's make sure we decimate every array properly...
        decim_list = {}
        for key, dset in estimator_datasets.items():
            dset_shape = list(dset.shape)
            n_slices = dset_shape[0] // stride
            dset_shape[0] = n_slices
            decim_set = np.empty((dset_shape), dtype=dset.dtype)
            for iout, istart in enumerate(range(0, dset.shape[0] - stride + 1, stride)):
                sl = dset[istart : istart + stride]
                # We assume time is the 0th axis.
                # Okay, so non-optimal.  Population requires the axis subsampling to be done just so...
                try:
                    decim_set[iout] = subsample(sl, axis=0)
                except Exception:
                    decim_set[iout] = subsample(sl)
            decim_list[key] = decim_set
            dlen = dset_shape[0]
            estimator_kwargs['stride'] = stride

        return mcbs_ci(
            dataset=decim_list,
            estimator=estimator,
            alpha=alpha,
            dlen=dlen,
            n_sets=n_sets,
            args=args,
            kwargs=estimator_kwargs,
            sort=np.msort,
        ) + (correl_len,)


# These are blocks designed to evaluate simple information sets.
# Whether they should go here or in westtoools is somewhat up for debate.
# Currently, nothing actually uses them, so there's that.


def _1D_simple_eval_block(
    iblock,
    start,
    stop,
    nstates,
    data_input,
    name,
    mcbs_alpha,
    mcbs_nsets,
    mcbs_acalpha,
    do_correl,
    mcbs_enable,
    subsample=np.mean,
    **extra
):
    # This is actually appropriate for anything with a directly measured, 1D dataset, i.e.,
    # Fluxes, color populations, and state populations.
    results = []
    for istate in range(nstates):
        # Not sure if we need a jstate for these estimators, but we'll see.
        # kwargs = {'istate': istate, 'jstate': 'B'}
        estimator_datasets = {'dataset': data_input['dataset'][:, istate]}
        ci_res = mcbs_ci_correl(
            estimator_datasets,
            estimator=(lambda stride, dataset: np.mean(dataset)),
            alpha=mcbs_alpha,
            n_sets=mcbs_nsets,
            autocorrel_alpha=mcbs_acalpha,
            subsample=subsample,
            do_correl=do_correl,
            mcbs_enable=mcbs_enable,
        )

        results.append((name, iblock, istate, (start, stop) + ci_res))

    return results


def _2D_simple_eval_block(
    iblock,
    start,
    stop,
    nstates,
    data_input,
    name,
    mcbs_alpha,
    mcbs_nsets,
    mcbs_acalpha,
    do_correl,
    mcbs_enable,
    subsample=np.mean,
    **extra
):
    # This is really just a simple 2D block for less complex datasets, but there it is.
    # It's probably limited in this use case to conditional_fluxes, but anything that's an i to j process that is directly measured
    # is suitable for use with this.
    results = []
    for istate in range(nstates):
        for jstate in range(nstates):
            if istate == jstate:
                continue
            # kwargs = {'istate': istate, 'jstate': jstate}
            # dataset = {'dataset': cond_fluxes[:, istate, jstate]}
            estimator_datasets = {'dataset': data_input['dataset'][:, istate, jstate]}
            ci_res = mcbs_ci_correl(
                estimator_datasets,
                estimator=(lambda stride, dataset: np.mean(dataset)),
                alpha=mcbs_alpha,
                n_sets=mcbs_nsets,
                autocorrel_alpha=mcbs_acalpha,
                subsample=subsample,
                do_correl=do_correl,
                mcbs_enable=mcbs_enable,
            )

            results.append((name, iblock, istate, jstate, (start, stop) + ci_res))

    return results




from __future__ import division
import numpy, sys
cimport numpy, cython
from cpython.buffer cimport *
from cpython.mem cimport *
from numpy cimport * #PyArray_DATA, PyArray_TYPE

ctypedef fused real_numeric:
    numpy.int8_t
    numpy.int16_t
    numpy.int32_t
    numpy.int64_t
    numpy.uint8_t
    numpy.uint16_t
    numpy.uint32_t
    numpy.uint64_t    
    numpy.float32_t
    numpy.float64_t

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef histnd(values, binbounds, weights=1.0, out=None, binbound_check = True, ignore_out_of_range=False):
    '''Generate an N-dimensional PDF (or contribution to a PDF) from the given values.
    ``binbounds`` is a list of arrays of boundary values, with one entry for each
    dimension (``values`` must have as many columns as there are entries in ``binbounds``)
    ``weight``, if provided, specifies the weight each value contributes to the
    histogram; this may be a scalar (for equal weights for all values) or a vector of
    the same length as ``values`` (for unequal weights). If ``binbound_check`` is True, then
    the boundaries are checked for strict positive monotonicity; set to False to shave a few
    microseconds if you know your bin boundaries to be monotonically increasing. 
    '''
    
    if values.ndim != 2:
        values = numpy.atleast_2d(values)
        if values.ndim > 2:
            raise TypeError('values must be 2-D')
    
    cdef:
        Py_ssize_t npts = values.shape[0]
        Py_ssize_t ndim = values.shape[1] 
        int typecode = PyArray_TYPE(values) 
        
    if len(binbounds) != ndim:
        raise ValueError('number of sets of bin boundaries ({}) does not match dimensionality of data ({})'
                         .format(len(binbounds), values.shape[1]))

    if binbound_check:
        for idim in xrange(ndim):
            dq = numpy.diff(binbounds[idim])
            if (dq <= 0).any():
                raise ValueError('binbounds in dimension {} are not strictly monotonically increasing'.format(idim))

    # Prepare bin boundaries arrays
    _binbounds_vectors = numpy.empty((ndim,), numpy.object_)
    _nbounds   = numpy.empty((ndim,), numpy.uint32)
    for idim in range(ndim):
        _binbounds = numpy.require(binbounds[idim], values.dtype, 'C')
        _binbounds_vectors[idim] = _binbounds
        _nbounds[idim] = _binbounds.shape[0]

    # Prepare output array, if necessary        
    if out is None:
        _out = numpy.zeros([len(boundset)-1 for boundset in binbounds], numpy.float64)
    else:
        _out = out
        if _out.dtype != numpy.float64:
            raise TypeError('type of output array must be float64')
        if not _out.flags.writeable:
            raise TypeError('output is not writeable') 
        
    # Prepare weight array
    _weights = numpy.require(weights, numpy.float64, 'C')
    if _weights.shape == ():
        # scalar
        _weights = numpy.empty((len(values),), numpy.float64)
        _weights[:] = weights
    elif _weights.ndim > 1:
        raise TypeError('weight must be scalar or one dimensional')
    elif  _weights.shape[0] != values.shape[0]:
        raise TypeError('weights and values must be equal in length')
    
    # ugh
    if typecode == NPY_FLOAT32:      
        return _histnd[numpy.float32_t](values,
                                        _binbounds_vectors,
                                        <numpy.uint32_t*> PyArray_DATA(_nbounds),
                                        <numpy.float64_t*> PyArray_DATA(_weights),
                                        ignore_out_of_range,
                                        _out)
    elif typecode == NPY_FLOAT64:
        return _histnd[numpy.float64_t](values,
                                        _binbounds_vectors,
                                        <numpy.uint32_t*> PyArray_DATA(_nbounds),
                                        <numpy.float64_t*> PyArray_DATA(_weights),
                                        ignore_out_of_range,
                                        _out)
    elif typecode == NPY_INT8:
        return _histnd[numpy.int8_t](values,
                                        _binbounds_vectors,
                                        <numpy.uint32_t*> PyArray_DATA(_nbounds),
                                        <numpy.float64_t*> PyArray_DATA(_weights),
                                        ignore_out_of_range,
                                        _out)
    elif typecode == NPY_INT16:
        return _histnd[numpy.int16_t](values,
                                        _binbounds_vectors,
                                        <numpy.uint32_t*> PyArray_DATA(_nbounds),
                                        <numpy.float64_t*> PyArray_DATA(_weights),
                                        ignore_out_of_range,
                                        _out)
    elif typecode == NPY_INT32:
        return _histnd[numpy.int32_t](values,
                                        _binbounds_vectors,
                                        <numpy.uint32_t*> PyArray_DATA(_nbounds),
                                        <numpy.float64_t*> PyArray_DATA(_weights),
                                        ignore_out_of_range,
                                        _out)
    elif typecode == NPY_INT64:
        return _histnd[numpy.int64_t](values,
                                        _binbounds_vectors,
                                        <numpy.uint32_t*> PyArray_DATA(_nbounds),
                                        <numpy.float64_t*> PyArray_DATA(_weights),
                                        ignore_out_of_range,
                                        _out)
    elif typecode == NPY_UINT8:
        return _histnd[numpy.uint8_t](values,
                                        _binbounds_vectors,
                                        <numpy.uint32_t*> PyArray_DATA(_nbounds),
                                        <numpy.float64_t*> PyArray_DATA(_weights),
                                        ignore_out_of_range,
                                        _out)
    elif typecode == NPY_UINT16:
        return _histnd[numpy.uint16_t](values,
                                        _binbounds_vectors,
                                        <numpy.uint32_t*> PyArray_DATA(_nbounds),
                                        <numpy.float64_t*> PyArray_DATA(_weights),
                                        ignore_out_of_range,
                                        _out)
    elif typecode == NPY_UINT32:
        return _histnd[numpy.uint32_t](values,
                                        _binbounds_vectors,
                                        <numpy.uint32_t*> PyArray_DATA(_nbounds),
                                        <numpy.float64_t*> PyArray_DATA(_weights),
                                        ignore_out_of_range,
                                        _out)
    elif typecode == NPY_UINT64:
        return _histnd[numpy.uint64_t](values,
                                        _binbounds_vectors,
                                        <numpy.uint32_t*> PyArray_DATA(_nbounds),
                                        <numpy.float64_t*> PyArray_DATA(_weights),
                                        ignore_out_of_range,
                                        _out)
    else:
        raise TypeError('real floating-point or integer input required')
    
@cython.boundscheck(False)
@cython.wraparound(False)
cdef _histnd(real_numeric[:,:] values, object[:] binbounds, numpy.uint32_t* nbounds, double* weights,
             bint ignore_out_of_range, object output):
    '''Bin the values stored in the 2-D array ``values`` with corresponding weights ``weights``
    into the array of bins ``output``. The bin boundaries are specified in ``binbounds``, and the
    length of each set of boundaries is stored in ``nbounds``.
    
    Pre-conditions:
     * output is backed by 64-bit floating point storage
     * len(binbounds) == len(nbounds) == values.ndim
     '''
     
    cdef:
        Py_buffer outputview, bbview
        Py_ssize_t ndim = values.shape[1], npts = values.shape[0]
        Py_ssize_t idim, ipt, ibound
        char* outptr_bytes
        double* outptr
        real_numeric val, lb, ub
        real_numeric* boundbuf
        real_numeric** _binbounds
        bint store_value

    # Get pointers to the (contiguous) lists of bin boundaries in each dimension
    _binbounds = <real_numeric**> PyMem_Malloc(ndim*sizeof(real_numeric*))
    if not _binbounds:
        raise MemoryError()
    for idim in range(ndim):
        PyObject_GetBuffer(binbounds[idim], &bbview, PyBUF_SIMPLE)
        _binbounds[idim] = <real_numeric*> bbview.buf
        PyBuffer_Release(&bbview)

    # Get a view of our output array, so we can write directly into it 
    PyObject_GetBuffer(output, &outputview, PyBUF_STRIDED)

    try:
        with nogil:
            # loop over points
            for ipt in range(npts):
                outptr_bytes = <char*> outputview.buf
                store_value = True
                for idim in range(ndim):
                    val = values[ipt,idim]
                    for ibound in range(nbounds[idim]-1):
                        lb = _binbounds[idim][ibound]
                        ub = _binbounds[idim][ibound+1]
                        if val >= lb and val < ub:
                            outptr_bytes += outputview.strides[idim] * ibound
                            break
                    else:
                        if not ignore_out_of_range:
                            with gil:
                                raise ValueError('value {} at index {} out of bin boundaries in dimension {}'
                                                 .format(val,ipt,idim))
                        else:
                            store_value = False
                if store_value:
                    outptr = <double*> outptr_bytes
                    outptr[0] += weights[ipt]
        return output
    finally:
        PyMem_Free(_binbounds)
        PyBuffer_Release(&outputview)





'''
Bin assignment for WEST simulations. This module defines "bin mappers" which take
vectors of coordinates (or rather, coordinate tuples), and assign each a definite
integer value identifying a bin. Critical portions are implemented in a Cython
extension module.

A number of pre-defined bin mappers are available here:

  * :class:`RectilinearBinMapper`, for bins divided by N-dimensional grids
  * :class:`FuncBinMapper`, for functions which directly calculate bin assignments
    for a number of coordinate values. This is best used with C/Cython/Numba
    functions, or intellegently-tuned numpy-based Python functions.
  * :class:`VectorizingFuncBinMapper`, for functions which calculate a bin
    assignment for a single coordinate value. This is best used for arbitrary
    Python functions.
  * :class:`PiecewiseBinMapper`, for using a set of boolean-valued functions, one
    per bin, to determine assignments. This is likely to be much slower than a
    `FuncBinMapper` or `VectorizingFuncBinMapper` equipped with an appropriate
    function, and its use is discouraged.

One "super-mapper" is available, for assembling more complex bin spaces from
simpler components:

  * :class:`RecursiveBinMapper`, for nesting one set of bins within another.

Users are also free to implement their own mappers. A bin mapper must implement, at
least, an ``assign(coords, mask=None, output=None)`` method, which is responsible
for mapping each of the vector of coordinate tuples ``coords`` to an integer
(np.uint16) indicating a what bin that coordinate tuple falls into. The optional
``mask`` (a numpy bool array) specifies that some coordinates are to be skipped; this is used,
for instance, by the recursive (nested) bin mapper to minimize the number of calculations
required to definitively assign a coordinate tuple to a bin. Similarly, the optional
``output`` must be an integer (uint16) array of the same length as ``coords``, into which
assignments are written. The ``assign()`` function must return a reference to ``output``.
(This is used to avoid allocating many temporary output arrays in complex binning
scenarios.)

A user-defined bin mapper must also make an ``nbins`` property available, containing
the total number of bins within the mapper.

'''

import hashlib
import logging
import pickle

import numpy as np

from .bins import Bin
from ._assign import output_map, apply_down, apply_down_argmin_across, rectilinear_assign

# All bin numbers are 16-bit unsigned ints, with one element (65525) reserved to
# indicate unknown or unassigned points. This allows up to 65,536 bins, making
# rate and flux matrices up to 32 GB (2**32 elements * 8 bytes). If you need more
# bins, change index_dtype here and index_dtype and index_t in _assign.pyx.
index_dtype = np.uint16
UNKNOWN_INDEX = 65535

# All coordinates are currently 32-bit floats. If you need 64-bit, change
# coord_dtype here and coord_t in _assign.pyx.
coord_dtype = np.float32


log = logging.getLogger(__name__)


class BinMapper:
    hashfunc = hashlib.sha256

    def __init__(self):
        self.labels = None
        self.nbins = 0

    def construct_bins(self, type_=Bin):
        '''Construct and return an array of bins of type ``type``'''
        return np.array([type_() for _i in range(self.nbins)], dtype=np.object_)

    def pickle_and_hash(self):
        '''Pickle this mapper and calculate a hash of the result (thus identifying the
        contents of the pickled data), returning a tuple ``(pickled_data, hash)``.
        This will raise PickleError if this mapper cannot be pickled, in which case
        code that would otherwise rely on detecting a topology change must assume
        a topology change happened, even if one did not.
        '''

        pkldat = pickle.dumps(self, pickle.HIGHEST_PROTOCOL)
        hash = self.hashfunc(pkldat)
        return (pkldat, hash.hexdigest())

    def __repr__(self):
        return '<{} at 0x{:x} with {:d} bins>'.format(self.__class__.__name__, id(self), self.nbins or 0)


class NopMapper(BinMapper):
    '''Put everything into one bin.'''

    def __init__(self):
        super().__init__()
        self.nbins = 1
        self.labels = ['nop']

    def assign(self, coords, mask=None, output=None):
        if output is None:
            output = np.zeros((len(coords),), dtype=index_dtype)

        if mask is None:
            mask = np.ones((len(coords),), dtype=np.bool_)
        else:
            mask = np.require(mask, dtype=np.bool_)

        output[mask] = 0


class RectilinearBinMapper(BinMapper):
    '''Bin into a rectangular grid based on tuples of float values'''

    def __init__(self, boundaries):
        super().__init__()
        self._boundaries = None
        self._boundlens = None
        self.ndim = 0
        self.nbins = 0

        # the setter function below handles all of the required wrangling
        self.boundaries = boundaries

    @property
    def boundaries(self):
        return self._boundaries

    @boundaries.setter
    def boundaries(self, boundaries):
        del self._boundaries, self.labels
        self._boundaries = []
        self.labels = labels = []
        for boundset in boundaries:
            boundarray = np.asarray(boundset, dtype=coord_dtype, order='C')
            db = np.diff(boundarray)
            if (db <= 0).any():
                raise ValueError('boundary set must be strictly monotonically increasing')
            self._boundaries.append(boundarray)
        self._boundlens = np.array([len(boundset) for boundset in self._boundaries], dtype=index_dtype)
        self.ndim = len(self._boundaries)
        self.nbins = np.multiply.accumulate([1] + [len(bounds) - 1 for bounds in self._boundaries])[-1]

        _boundaries = self._boundaries
        binspace_shape = tuple(self._boundlens[:] - 1)
        for index in np.ndindex(binspace_shape):
            bounds = [(_boundaries[idim][index[idim]], boundaries[idim][index[idim] + 1]) for idim in range(len(_boundaries))]
            labels.append(repr(bounds))

    def assign(self, coords, mask=None, output=None):
        try:
            passed_coord_dtype = coords.dtype
        except AttributeError:
            coords = np.require(coords, dtype=coord_dtype)
        else:
            if passed_coord_dtype != coord_dtype:
                coords = np.require(coords, dtype=coord_dtype)

        if coords.ndim != 2:
            raise TypeError('coords must be 2-dimensional')

        if mask is None:
            mask = np.ones((len(coords),), dtype=np.bool_)
        elif len(mask) != len(coords):
            raise TypeError('mask [shape {}] has different length than coords [shape {}]'.format(mask.shape, coords.shape))

        if output is None:
            output = np.empty((len(coords),), dtype=index_dtype)
        elif len(output) != len(coords):
            raise TypeError('output has different length than coords')

        rectilinear_assign(coords, mask, output, self.boundaries, self._boundlens)

        return output


class PiecewiseBinMapper(BinMapper):
    '''Binning using a set of functions returing boolean values; if the Nth function
    returns True for a coordinate tuple, then that coordinate is in the Nth bin.'''

    def __init__(self, functions):
        self.functions = functions
        self.nbins = len(functions)
        self.index_dtype = np.min_scalar_type(self.nbins)
        self.labels = [repr(func) for func in functions]

    def assign(self, coords, mask=None, output=None):
        if output is None:
            output = np.zeros((len(coords),), dtype=index_dtype)

        if mask is None:
            mask = np.ones((len(coords),), dtype=np.bool_)
        else:
            mask = np.require(mask, dtype=np.bool_)

        coord_subset = coords[mask]
        fnvals = np.empty((len(coord_subset), len(self.functions)), dtype=index_dtype)
        for ifn, fn in enumerate(self.functions):
            rsl = np.apply_along_axis(fn, 0, coord_subset)
            if rsl.ndim > 1:
                # this should work like a squeeze, unless the function returned something truly
                # stupid (e.g., a 3d array with at least two dimensions greater than 1), in which
                # case a broadcast error will occur
                fnvals[:, ifn] = rsl.flat
            else:
                fnvals[:, ifn] = rsl
        amask = np.require(fnvals.argmax(axis=1), dtype=index_dtype)
        output[mask] = amask
        return output


class FuncBinMapper(BinMapper):
    '''Binning using a custom function which must iterate over input coordinate
    sets itself.'''

    def __init__(self, func, nbins, args=None, kwargs=None):
        self.func = func
        self.nbins = nbins
        self.args = args or ()
        self.kwargs = kwargs or {}
        self.labels = ['{!r} bin {:d}'.format(func, ibin) for ibin in range(nbins)]

    def assign(self, coords, mask=None, output=None):
        try:
            passed_coord_dtype = coords.dtype
        except AttributeError:
            coords = np.require(coords, dtype=coord_dtype)
        else:
            if passed_coord_dtype != coord_dtype:
                coords = np.require(coords, dtype=coord_dtype)

        if coords.ndim != 2:
            raise TypeError('coords must be 2-dimensional')
        if mask is None:
            mask = np.ones((len(coords),), dtype=np.bool_)
        elif len(mask) != len(coords):
            raise TypeError('mask [shape {}] has different length than coords [shape {}]'.format(mask.shape, coords.shape))

        if output is None:
            output = np.empty((len(coords),), dtype=index_dtype)
        elif len(output) != len(coords):
            raise TypeError('output has different length than coords')

        self.func(coords, mask, output, *self.args, **self.kwargs)

        return output


class VectorizingFuncBinMapper(BinMapper):
    '''Binning using a custom function which is evaluated once for each (unmasked)
    coordinate tuple provided.'''

    def __init__(self, func, nbins, args=None, kwargs=None):
        self.func = func
        self.args = args or ()
        self.kwargs = kwargs or {}
        self.nbins = nbins
        self.index_dtype = np.min_scalar_type(self.nbins)
        self.labels = ['{!r} bin {:d}'.format(func, ibin) for ibin in range(nbins)]

    def assign(self, coords, mask=None, output=None):
        try:
            passed_coord_dtype = coords.dtype
        except AttributeError:
            coords = np.require(coords, dtype=coord_dtype)
        else:
            if passed_coord_dtype != coord_dtype:
                coords = np.require(coords, dtype=coord_dtype)

        if coords.ndim != 2:
            raise TypeError('coords must be 2-dimensional')
        if mask is None:
            mask = np.ones((len(coords),), dtype=np.bool_)
        elif len(mask) != len(coords):
            raise TypeError('mask [shape {}] has different length than coords [shape {}]'.format(mask.shape, coords.shape))

        if output is None:
            output = np.empty((len(coords),), dtype=index_dtype)
        elif len(output) != len(coords):
            raise TypeError('output has different length than coords')

        apply_down(self.func, self.args, self.kwargs, coords, mask, output)

        return output


class VoronoiBinMapper(BinMapper):
    '''A one-dimensional mapper which assigns a multidimensional pcoord to the
    closest center based on a distance metric. Both the list of centers and the
    distance function must be supplied.'''

    def __init__(self, dfunc, centers, dfargs=None, dfkwargs=None):
        self.dfunc = dfunc
        self.dfargs = dfargs or ()
        self.dfkwargs = dfkwargs or {}
        self.centers = np.asarray(centers)
        self.nbins = self.centers.shape[0]
        self.ndim = self.centers.shape[1]
        self.labels = ['center={!r}'.format(center) for center in self.centers]

        # Sanity check: does the distance map the centers to themselves?
        check = self.assign(self.centers)
        if (check != np.arange(len(self.centers))).any():
            raise TypeError('dfunc does not map centers to themselves')

    def assign(self, coords, mask=None, output=None):
        try:
            passed_coord_dtype = coords.dtype
        except AttributeError:
            coords = np.require(coords, dtype=coord_dtype)
        else:
            if passed_coord_dtype != coord_dtype:
                coords = np.require(coords, dtype=coord_dtype)

        if coords.ndim != 2:
            raise TypeError('coords must be 2-dimensional')
        if mask is None:
            mask = np.ones((len(coords),), dtype=np.bool_)
        elif len(mask) != len(coords):
            raise TypeError('mask [shape {}] has different length than coords [shape {}]'.format(mask.shape, coords.shape))

        if output is None:
            output = np.empty((len(coords),), dtype=index_dtype)
        elif len(output) != len(coords):
            raise TypeError('output has different length than coords')

        apply_down_argmin_across(self.dfunc, (self.centers,) + self.dfargs, self.dfkwargs, self.nbins, coords, mask, output)

        return output


class RecursiveBinMapper(BinMapper):
    '''Nest mappers one within another.'''

    def __init__(self, base_mapper, start_index=0):
        self.base_mapper = base_mapper
        self.nbins = base_mapper.nbins

        # Targets for recursion
        self._recursion_targets = {}

        # Which bins must we recurse into?
        self._recursion_map = np.zeros((self.base_mapper.nbins,), dtype=np.bool_)

        self.start_index = start_index

    @property
    def labels(self):
        for ilabel in range(self.base_mapper.nbins):
            if self._recursion_map[ilabel]:
                for label in self._recursion_targets[ilabel].labels:
                    yield label
            else:
                yield self.base_mapper.labels[ilabel]

    @property
    def start_index(self):
        return self._start_index

    @start_index.setter
    def start_index(self, new_index):
        self._start_index = new_index
        not_recursed = ~self._recursion_map
        n_not_recursed = not_recursed.sum()
        if n_not_recursed == self.nbins:
            self._output_map = np.arange(self._start_index, self._start_index + self.nbins, dtype=index_dtype)
        elif n_not_recursed > 0:
            # This looks like uninitialized access, but self._output_map is always set during __init__
            # (by self.start_index = 0, or whatever value was passed in), so this modifies the existing
            # set chosen above
            self._output_map[not_recursed] = np.arange(self._start_index, self._start_index + n_not_recursed, dtype=index_dtype)
        else:
            # No un-replaced bins
            self._output_map = None

        n_own_bins = self.base_mapper.nbins - self._recursion_map.sum()
        startindex = self.start_index + n_own_bins
        for mapper in self._recursion_targets.values():
            mapper.start_index = startindex
            startindex += mapper.nbins

    def add_mapper(self, mapper, replaces_bin_at):
        '''Replace the bin containing the coordinate tuple ``replaces_bin_at`` with the
        specified ``mapper``.'''

        replaces_bin_at = np.require(replaces_bin_at, dtype=coord_dtype)
        if replaces_bin_at.ndim < 1:
            replaces_bin_at.shape = (1, 1)
        elif replaces_bin_at.ndim < 2:
            replaces_bin_at.shape = (1, replaces_bin_at.shape[0])
        elif replaces_bin_at.ndim > 2 or replaces_bin_at.shape[1] > 1:
            raise TypeError('a single coordinate vector is required')

        self.nbins += mapper.nbins - 1

        ibin = self.base_mapper.assign(replaces_bin_at)[0]
        log.debug('replacing bin {!r} containing {!r} with {!r}'.format(ibin, replaces_bin_at, mapper))
        if self._recursion_map[ibin]:
            # recursively add; this doesn't change anything for us except our
            # total bin count, which has been accounted for above
            self._recursion_targets[ibin].add_mapper(mapper, replaces_bin_at[0])
        else:
            # replace a bin on our mapper
            self._recursion_map[ibin] = True
            mapper = RecursiveBinMapper(mapper)
            self._recursion_targets[ibin] = mapper

        # we have updated our list of recursed bins, so set our own start index to trigger a recursive
        # reassignment of mappers' output values
        # Note that we're reordering the recursion targets based on outer bin numbers (dict keys) first,
        # so the order the mappers were added no longer matters...
        self._recursion_targets = {k: self._recursion_targets[k] for k in sorted(self._recursion_targets)}
        self.start_index = self.start_index

    def assign(self, coords, mask=None, output=None):
        if mask is None:
            mask = np.ones((len(coords),), dtype=np.bool_)

        if output is None:
            output = np.empty((len(coords),), dtype=index_dtype)

        # mapping mask -- which output values come from our base
        # region set and therefore must be remapped
        mmask = np.zeros((len(coords),), dtype=np.bool_)

        # Assign based on this mapper
        self.base_mapper.assign(coords, mask, output)

        # Which coordinates do we need to reassign, because they landed in
        # bins with embedded mappers?
        rmasks = {}
        for rindex, mapper in self._recursion_targets.items():
            omask = output == rindex
            mmask |= omask
            rmasks[rindex] = omask

        # remap output from our (base) mapper
        # omap may be None if every bin has a recursive mapper in it
        omap = self._output_map
        if omap is not None:
            output_map(output, omap, mask & ~mmask)

        # do any recursive assignments necessary
        for rindex, mapper in self._recursion_targets.items():
            mapper.assign(coords, mask & rmasks[rindex], output)

        return output



import numpy as np
from numpy import index_exp

from .core import WESTToolComponent

import westpa
from westpa.core.extloader import get_object
from westpa.core.h5io import FnDSSpec, MultiDSSpec, SingleSegmentDSSpec, SingleIterDSSpec


def _get_parent_ids(n_iter, iter_group):
    seg_index = iter_group['seg_index']
    try:
        return seg_index['parent_id'][:]
    except ValueError:
        # field not found
        offsets = seg_index['parents_offset'][:]
        all_parents = iter_group['parents'][...]
        return np.require(all_parents.take(offsets), dtype=np.int64)
    else:
        return seg_index['parent_id']


class WESTDataReader(WESTToolComponent):
    '''Tool for reading data from WEST-related HDF5 files. Coordinates finding
    the main HDF5 file from west.cfg or command line arguments, caching of certain
    kinds of data (eventually), and retrieving auxiliary data sets from various
    places.'''

    def __init__(self):
        super().__init__()
        self.data_manager = westpa.rc.get_data_manager()
        self.we_h5filename = None

        self._weight_dsspec = None
        self._parent_id_dsspec = None

    def add_args(self, parser):
        group = parser.add_argument_group('WEST input data options')
        group.add_argument(
            '-W',
            '--west-data',
            dest='we_h5filename',
            metavar='WEST_H5FILE',
            help='''Take WEST data from WEST_H5FILE (default: read from the HDF5 file specified in west.cfg).''',
        )

    def process_args(self, args):
        if args.we_h5filename:
            self.data_manager.we_h5filename = self.we_h5filename = args.we_h5filename
        else:
            self.we_h5filename = self.data_manager.we_h5filename

    def open(self, mode='r'):
        self.data_manager.open_backing(mode)

    def close(self):
        self.data_manager.close_backing()

    def __getattr__(self, key):
        return getattr(self.data_manager, key)

    @property
    def weight_dsspec(self):
        if self._weight_dsspec is None:
            assert self.we_h5filename is not None
            self._weight_dsspec = SingleIterDSSpec(self.we_h5filename, 'seg_index', slice=index_exp['weight'])
        return self._weight_dsspec

    @property
    def parent_id_dsspec(self):
        if self._parent_id_dsspec is None:
            assert self.we_h5filename is not None
            # self._parent_id_dsspec = SingleIterDSSpec(self.we_h5filename, 'seg_index', slice=index_exp['parent_id'])
            self._parent_id_dsspec = FnDSSpec(self.we_h5filename, _get_parent_ids)
        return self._parent_id_dsspec

    def __enter__(self):
        self.open('r')
        return self

    def __exit__(self, exc_type, exc_val, exc_traceback):
        self.close()
        return False


class WESTDSSynthesizer(WESTToolComponent):
    '''Tool for synthesizing a dataset for analysis from other datasets. This
    may be done using a custom function, or a list of "data set specifications".
    It is anticipated that if several source datasets are required, then a tool
    will have multiple instances of this class.'''

    group_name = 'input dataset options'

    def __init__(self, default_dsname=None, h5filename=None):
        super().__init__()

        self.h5filename = h5filename
        self.default_dsname = default_dsname

        self.dsspec = None

    def add_args(self, parser):
        igroup = parser.add_argument_group(self.group_name).add_mutually_exclusive_group(required=not bool(self.default_dsname))

        igroup.add_argument(
            '--construct-dataset',
            help='''Use the given function (as in module.function) to extract source data.
                            This function will be called once per iteration as function(n_iter, iter_group)
                            to construct data for one iteration. Data returned must be indexable as
                            [seg_id][timepoint][dimension]''',
        )

        igroup.add_argument('--dsspecs', nargs='+', metavar='DSSPEC', help='''Construct source data from one or more DSSPECs.''')

    def process_args(self, args):
        if args.construct_dataset:
            self.dsspec = FnDSSpec(self.h5filename, get_object(args.construct_dataset, path=['.']))
        elif args.dsspecs:
            self.dsspec = MultiDSSpec([SingleSegmentDSSpec.from_string(dsspec, self.h5filename) for dsspec in args.dsspecs])
        else:
            # we can only get here if a default dataset name was specified
            assert self.default_dsname
            self.dsspec = SingleSegmentDSSpec(self.h5filename, self.default_dsname)


class WESTWDSSynthesizer(WESTToolComponent):
    group_name = 'weight dataset options'

    def __init__(self, default_dsname=None, h5filename=None):
        super(WESTWDSSynthesizer, self).__init__()

        self.h5filename = h5filename
        self.default_dsname = default_dsname

        self.dsspec = None

    def add_args(self, parser):
        wgroup = parser.add_argument_group(self.group_name).add_mutually_exclusive_group(required=not bool(self.default_dsname))

        wgroup.add_argument(
            '--construct-wdataset',
            help='''Use the given function (as in module.function) to extract source data.
                            This function will be called once per iteration as function(n_iter, iter_group)
                            to construct data for one iteration. Data returned must be indexable as
                            [seg_id]''',
        )
        wgroup.add_argument('--wdsspecs', nargs='+', metavar='WDSSPEC', help='''Construct weight data from one or more DSSPECs.''')

    def process_args(self, args):
        if args.construct_wdataset:
            self.dsspec = FnDSSpec(self.h5filename, get_object(args.construct_wdataset, path=['.']))
        elif args.dsspecs:
            self.dsspec = MultiDSSpec([SingleSegmentDSSpec.from_string(dsspec, self.h5filename) for dsspec in args.dsspecs])
        else:
            # we can only get here if a default dataset name was specified
            assert self.default_dsname
            # we gotta slice by weight for weights if we want to get the default to work
            self.dsspec = SingleIterDSSpec(self.h5filename, self.default_dsname, slice=np.index_exp['weight'])



'''
Created on May 29, 2015

@author: mzwier
'''

import logging
import multiprocessing
import os
import signal
import threading

from .core import ZMQCore, Message, ZMQWMTimeout, PassiveMultiTimer, Task, Result, TIMEOUT_MASTER_BEACON

import zmq


log = logging.getLogger(__name__)


class ZMQWorker(ZMQCore):
    '''This is the outward facing worker component of the ZMQ work manager. This
    forms the interface to the master. This process cannot hang or crash due to an
    error in tasks it executes, so tasks are isolated in ZMQExecutor, which
    communicates with ZMQWorker via (what else?) ZeroMQ.'''

    def __init__(self, rr_endpoint, ann_endpoint):
        super().__init__()

        # Upstream endpoints
        self.rr_endpoint = rr_endpoint
        self.ann_endpoint = ann_endpoint

        # Downstream endpoints
        self.task_endpoint = self.make_internal_endpoint()
        self.result_endpoint = self.make_internal_endpoint()

        self.master_id = None
        self.identified = False

        # The task currently being processed
        self.pending_task = None

        # Executor process

        self.shutdown_timeout = 5.0  # Five second wait between shutdown message and SIGINT and SIGINT and SIGKILL
        self.executor_process = None

    @property
    def is_master(self):
        return False

    def update_master_info(self, msg):
        if self.master_id is None:
            self.master_id = msg.master_id
        self.timers.reset(TIMEOUT_MASTER_BEACON)

    def identify(self, rr_socket):
        if self.master_id is None or self.identified or self.timers.expired(TIMEOUT_MASTER_BEACON):
            return
        self.send_message(rr_socket, Message.IDENTIFY, payload=self.get_identification())
        self.recv_ack(rr_socket, timeout=self.master_beacon_period * self.timeout_factor * 1000)
        self.identified = True

    def request_task(self, rr_socket, task_socket):
        if self.master_id is None:
            return
        elif self.pending_task is not None:
            return
        elif self.timers.expired(TIMEOUT_MASTER_BEACON):
            return
        else:
            self.send_message(rr_socket, Message.TASK_REQUEST)
            reply = self.recv_message(rr_socket, timeout=self.master_beacon_period * self.timeout_factor * 1000)
            self.update_master_info(reply)
            if reply.message == Message.NAK:
                # No task available
                return
            else:
                with self.message_validation(reply):
                    assert isinstance(reply.payload, Task)
                    task = reply.payload
                self.pending_task = task
                self.send_message(task_socket, Message.TASK, task)

    def handle_reconfigure_timeout(self, msg, timers):
        with self.message_validation(msg):
            assert msg.payload is not None
            timer, new_period = msg.payload
        timers.change_duration(timer, new_period)
        timers.reset(timer)

    def handle_result(self, result_socket, rr_socket):
        msg = self.recv_message(result_socket)
        with self.message_validation(msg):
            assert msg.message == Message.RESULT
            assert isinstance(msg.payload, Result)
            assert msg.payload.task_id == self.pending_task.task_id

        msg.src_id = self.node_id
        self.pending_task = None
        self.send_message(rr_socket, msg)
        reply = self.recv_ack(rr_socket, timeout=self.master_beacon_period * self.timeout_factor * 1000)
        self.update_master_info(reply)

    def comm_loop(self):
        '''Master communication loop for the worker process.'''

        rr_socket = self.context.socket(zmq.REQ)

        ann_socket = self.context.socket(zmq.SUB)
        ann_socket.setsockopt(zmq.SUBSCRIBE, b'')
        inproc_socket = self.context.socket(zmq.SUB)
        inproc_socket.setsockopt(zmq.SUBSCRIBE, b'')

        task_socket = self.context.socket(zmq.PUSH)
        result_socket = self.context.socket(zmq.PULL)

        self.log.info('This is {}'.format(self.node_description))

        timers = self.timers = PassiveMultiTimer()
        timers.add_timer(TIMEOUT_MASTER_BEACON, 86400)
        timers.add_timer('worker_beacon', self.worker_beacon_period)
        timers.add_timer('startup_timeout', self.startup_timeout)
        peer_found = False

        try:
            rr_socket.connect(self.rr_endpoint)
            ann_socket.connect(self.ann_endpoint)
            inproc_socket.bind(self.inproc_endpoint)
            task_socket.connect(self.task_endpoint)
            result_socket.bind(self.result_endpoint)

            poller = zmq.Poller()
            poller.register(ann_socket, zmq.POLLIN)
            poller.register(inproc_socket, zmq.POLLIN)
            poller.register(result_socket, zmq.POLLIN)

            timers.reset()
            while True:
                # If a timer is already expired, next_expiration_in() will return 0, which
                # zeromq interprets as infinite wait; so instead we select a 1 ms wait in this
                # case.
                poll_results = dict(poller.poll((timers.next_expiration_in() or 0.001) * 1000))

                if poll_results and not peer_found:
                    timers.remove_timer('startup_timeout')
                    peer_found = True
                    timers.change_duration(TIMEOUT_MASTER_BEACON, self.master_beacon_period * self.timeout_factor)
                    timers.reset(TIMEOUT_MASTER_BEACON)

                announcements = []

                # Check for internal messages first
                if inproc_socket in poll_results:
                    announcements.extend(self.recv_all(inproc_socket))

                # Process announcements
                if ann_socket in poll_results:
                    announcements.extend(self.recv_all(ann_socket))

                # announcements = Message.coalesce_announcements(announcements)
                # self.log.debug('received {:d} announcements'.format(len(announcements)))

                messages_by_tag = {}
                for msg in announcements:
                    messages_by_tag.setdefault(msg.message, list()).append(msg)

                # Check for shutdown messages
                if Message.SHUTDOWN in messages_by_tag:
                    self.log.debug('received shutdown message')
                    return
                elif Message.TASKS_AVAILABLE in messages_by_tag:
                    self.update_master_info(messages_by_tag[Message.TASKS_AVAILABLE][0])
                elif Message.MASTER_BEACON in messages_by_tag:
                    self.update_master_info(messages_by_tag[Message.MASTER_BEACON][0])

                if self.master_id is not None and timers.expired('worker_beacon'):
                    self.identify(rr_socket)
                    timers.reset('worker_beacon')

                # Handle results, so that we clear ourselves of completed tasks
                # before asking for more
                if result_socket in poll_results:
                    self.handle_result(result_socket, rr_socket)
                    # immediately request another task if available
                    if not timers.expired(TIMEOUT_MASTER_BEACON):
                        self.request_task(rr_socket, task_socket)

                # Handle any remaining messages
                for tag, msgs in messages_by_tag.items():
                    if tag == Message.MASTER_BEACON:
                        self.identify(rr_socket)
                    elif tag == Message.RECONFIGURE_TIMEOUT:
                        for msg in msgs:
                            self.handle_reconfigure_timeout(msg, timers)
                    elif tag == Message.TASKS_AVAILABLE:
                        self.request_task(rr_socket, task_socket)

                del announcements, messages_by_tag

                if timers.expired(TIMEOUT_MASTER_BEACON):
                    self.log.error('no contact from master; shutting down')
                    return

                if not peer_found and timers.expired('startup_timeout'):
                    self.log.error('startup phase elapsed with no contact from master; shutting down')
                    break

        except ZMQWMTimeout:
            # both handle_result() and request_task() have receive timeouts set to
            # self.master_beacon_period*self.timeout_factor, and timeout exceptions
            # propagate to this point.
            self.log.error('timeout communicating with peer; shutting down')
        finally:
            self.shutdown_executor()
            self.executor_process.join()
            self.context.destroy(linger=1)
            self.context = None
            self.remove_ipc_endpoints()

    def shutdown_executor(self):
        if self.context is not None:
            try:
                self.log.debug('sending shutdown task to executor')
                task_socket = self.context.socket(zmq.PUSH)
                task_socket.connect(self.task_endpoint)
                self.send_message(task_socket, Message.SHUTDOWN)
                task_socket.close(linger=1)
            except Exception:
                pass

        pid = self.executor_process.pid
        self.executor_process.join(self.shutdown_timeout)
        # is_alive() is prone to a race condition so catch the case that the PID is already dead
        if self.executor_process.is_alive():
            self.log.debug('sending SIGTERM to worker process {:d}'.format(pid))
            self.executor_process.terminate()
            # try:
            #     os.kill(self.executor_process.pid, signal.SIGINT)
            # except ProcessLookupError:
            #     self.log.debug('worker process {:d} already dead'.format(pid))
            self.executor_process.join(self.shutdown_timeout)
            if self.executor_process.is_alive():
                self.executor_process.kill()
                self.log.warning('sending SIGKILL to worker process {:d}'.format(pid))
                # try:
                #     os.kill(self.executor_process.pid, signal.SIGKILL)
                # except ProcessLookupError:
                #     self.log.debug('worker process {:d} already dead'.format(pid))
            self.executor_process.join()
            self.log.debug('worker process {:d} terminated'.format(pid))
        else:
            self.log.debug('worker process {:d} terminated gracefully with code {:d}'.format(pid, self.executor_process.exitcode))

    def install_signal_handlers(self, signals=None):
        if not signals:
            signals = {signal.SIGINT, signal.SIGQUIT, signal.SIGTERM}

        for sig in signals:
            signal.signal(sig, signal.SIG_IGN)

    def startup(self, process_index=None):
        self.install_signal_handlers()
        executor = ZMQExecutor(self.task_endpoint, self.result_endpoint)
        self.executor_process = multiprocessing.Process(target=executor.startup, args=(process_index,))
        self.executor_process.start()
        self.context = zmq.Context()
        self.comm_thread = threading.Thread(target=self.comm_loop)
        self.comm_thread.start()


class ZMQExecutor(ZMQCore):
    '''The is the component of the ZMQ WM worker that actually executes tasks.
    This is isolated in a separate process and controlled via ZMQ from
    the ZMQWorker.'''

    def __init__(self, task_endpoint, result_endpoint):
        super().__init__()

        self.task_endpoint = task_endpoint
        self.result_endpoint = result_endpoint

    def comm_loop(self):
        task_socket = self.context.socket(zmq.PULL)
        result_socket = self.context.socket(zmq.PUSH)

        task_socket.bind(self.task_endpoint)
        result_socket.connect(self.result_endpoint)

        self.log.info('This is {}'.format(self.node_description))

        try:
            while True:
                try:
                    msg = self.recv_message(task_socket, timeout=100)
                except KeyboardInterrupt:
                    break
                except ZMQWMTimeout:
                    continue
                else:
                    if msg.message == Message.TASK:
                        task = msg.payload
                        result = task.execute()
                        self.send_message(result_socket, Message.RESULT, result)
                    elif msg.message == Message.SHUTDOWN:
                        break
        finally:
            self.context.destroy(linger=0)
            self.context = None

    def startup(self, process_index=None):
        if process_index is not None:
            from westpa.work_managers import environment

            pi_name = '{}_PROCESS_INDEX'.format(environment.WMEnvironment.env_prefix)
            self.log.debug('Setting {}={}'.format(pi_name, process_index))
            os.environ[pi_name] = str(process_index)
        self.context = zmq.Context()
        self.comm_loop()



import importlib
import logging
import sys

log = logging.getLogger(__name__)


def load_module(module_name, path=None):
    """Load and return the given module, recursively loading containing packages as necessary."""
    if module_name in sys.modules:
        log.debug('module %r already loaded' % module_name)
        return sys.modules[module_name]

    if path is None:
        return importlib.import_module(module_name)

    spec_components = list(reversed(module_name.split('.')))
    qname_components = []
    mod_chain = []
    while spec_components:
        next_component = spec_components.pop(-1)
        qname_components.append(next_component)

        try:
            parent = mod_chain[-1]
            path = parent.__path__
        except IndexError:
            parent = None

        qname = '.'.join(qname_components)

        if qname in sys.modules:
            log.debug("Skipping preloaded module {}".format(qname))
            module = sys.modules[qname]
        else:
            log.debug('find_spec({!r}, {!r})'.format(qname, path))

            spec = importlib.machinery.PathFinder().find_spec(qname, path)

            if spec is None:
                raise ImportError(f'No module named {qname}')

            module = importlib.util.module_from_spec(spec)

            if spec.name not in sys.modules:
                sys.modules[spec.name] = module

            spec.loader.exec_module(module)

            # Make the module appear in the parent module's namespace
            if parent:
                setattr(parent, next_component, module)

            log.debug('module %r loaded' % qname)

        mod_chain.append(module)

    return module


def get_object(object_name, path=None):
    """Attempt to load the given object, using additional path information if given."""

    try:
        (modspec, symbol) = object_name.rsplit('.', 1)
    except ValueError:
        # no period found
        raise ValueError("object_name name must be in the form 'module.symbol'")

    log.debug('attempting to load %r from %r' % (symbol, modspec))
    module = load_module(modspec, path)

    # This will raise AttributeError (as expected) if the symbol is not in the module
    return getattr(module, symbol)
